import json
from datetime import UTC, datetime
from typing import Any, cast

import structlog

from skyvern.config import settings
from skyvern.forge.sdk.db.enums import TaskType
from skyvern.forge.sdk.db.id import (
    generate_aws_secret_parameter_id,
    generate_azure_vault_credential_parameter_id,
    generate_bitwarden_credit_card_data_parameter_id,
    generate_bitwarden_login_credential_parameter_id,
    generate_bitwarden_sensitive_information_parameter_id,
    generate_credential_parameter_id,
    generate_onepassword_credential_parameter_id,
    generate_output_parameter_id,
    generate_workflow_parameter_id,
)
from skyvern.forge.sdk.workflow.exceptions import (
    ContextParameterSourceNotDefined,
    InvalidWaitBlockTime,
    InvalidWorkflowDefinition,
    WorkflowDefinitionHasDuplicateParameterKeys,
    WorkflowDefinitionHasReservedParameterKeys,
    WorkflowDefinitionHasUndefinedParameters,
    WorkflowParameterMissingRequiredValue,
)
from skyvern.forge.sdk.workflow.models.block import (
    ActionBlock,
    BlockTypeVar,
    BranchCondition,
    CodeBlock,
    ConditionalBlock,
    DownloadToS3Block,
    ExtractionBlock,
    FileDownloadBlock,
    FileParserBlock,
    FileUploadBlock,
    ForLoopBlock,
    HttpRequestBlock,
    HumanInteractionBlock,
    JinjaBranchCriteria,
    LoginBlock,
    NavigationBlock,
    PDFParserBlock,
    PrintPageBlock,
    PromptBranchCriteria,
    SendEmailBlock,
    TaskBlock,
    TaskV2Block,
    TextPromptBlock,
    UploadToS3Block,
    UrlBlock,
    ValidationBlock,
    WaitBlock,
)
from skyvern.forge.sdk.workflow.models.parameter import (
    PARAMETER_TYPE,
    RESERVED_PARAMETER_KEYS,
    AWSSecretParameter,
    AzureVaultCredentialParameter,
    BitwardenCreditCardDataParameter,
    BitwardenLoginCredentialParameter,
    BitwardenSensitiveInformationParameter,
    ContextParameter,
    CredentialParameter,
    OnePasswordCredentialParameter,
    OutputParameter,
    Parameter,
    ParameterType,
    WorkflowParameter,
    WorkflowParameterType,
)
from skyvern.forge.sdk.workflow.models.workflow import (
    WorkflowDefinition,
)
from skyvern.schemas.workflows import (
    BLOCK_YAML_TYPES,
    BlockType,
    ForLoopBlockYAML,
    WorkflowDefinitionYAML,
)

LOG = structlog.get_logger()


def convert_workflow_definition(
    workflow_definition_yaml: WorkflowDefinitionYAML,
    workflow_id: str,
) -> WorkflowDefinition:
    # Create parameters from the request
    parameters: dict[str, PARAMETER_TYPE] = {}
    duplicate_parameter_keys = set()

    # Check if user's trying to manually create an output parameter
    if any(parameter.parameter_type == ParameterType.OUTPUT for parameter in workflow_definition_yaml.parameters):
        raise InvalidWorkflowDefinition(message="Cannot manually create output parameters")

    # Check if any parameter keys collide with automatically created output parameter keys
    block_labels = [block.label for block in workflow_definition_yaml.blocks]
    # TODO (kerem): Check if block labels are unique
    output_parameter_keys = [f"{block_label}_output" for block_label in block_labels]
    parameter_keys = [parameter.key for parameter in workflow_definition_yaml.parameters]
    if any(key in output_parameter_keys for key in parameter_keys):
        raise WorkflowDefinitionHasReservedParameterKeys(
            reserved_keys=output_parameter_keys, parameter_keys=parameter_keys
        )

    if any(key in RESERVED_PARAMETER_KEYS for key in parameter_keys):
        raise WorkflowDefinitionHasReservedParameterKeys(
            reserved_keys=RESERVED_PARAMETER_KEYS,
            parameter_keys=parameter_keys,
        )

    # Create output parameters for all blocks
    block_output_parameters = _create_all_output_parameters_for_workflow(
        workflow_id=workflow_id,
        block_yamls=workflow_definition_yaml.blocks,
    )
    for block_output_parameter in block_output_parameters.values():
        parameters[block_output_parameter.key] = block_output_parameter

    # We're going to process context parameters after other parameters since they depend on the other parameters
    context_parameter_yamls = []

    for parameter in workflow_definition_yaml.parameters:
        if parameter.key in parameters:
            LOG.error(f"Duplicate parameter key {parameter.key}")
            duplicate_parameter_keys.add(parameter.key)
            continue
        now = datetime.now(UTC)
        if parameter.parameter_type == ParameterType.AWS_SECRET:
            parameters[parameter.key] = AWSSecretParameter(
                aws_secret_parameter_id=generate_aws_secret_parameter_id(),
                workflow_id=workflow_id,
                aws_key=parameter.aws_key,
                key=parameter.key,
                description=parameter.description,
                created_at=now,
                modified_at=now,
            )
        elif parameter.parameter_type == ParameterType.CREDENTIAL:
            parameters[parameter.key] = CredentialParameter(
                credential_parameter_id=generate_credential_parameter_id(),
                workflow_id=workflow_id,
                key=parameter.key,
                description=parameter.description,
                credential_id=parameter.credential_id,
                created_at=now,
                modified_at=now,
            )
        elif parameter.parameter_type == ParameterType.ONEPASSWORD:
            parameters[parameter.key] = OnePasswordCredentialParameter(
                onepassword_credential_parameter_id=generate_onepassword_credential_parameter_id(),
                workflow_id=workflow_id,
                key=parameter.key,
                description=parameter.description,
                vault_id=parameter.vault_id,
                item_id=parameter.item_id,
                created_at=now,
                modified_at=now,
            )
        elif parameter.parameter_type == ParameterType.AZURE_VAULT_CREDENTIAL:
            parameters[parameter.key] = AzureVaultCredentialParameter(
                azure_vault_credential_parameter_id=generate_azure_vault_credential_parameter_id(),
                workflow_id=workflow_id,
                key=parameter.key,
                description=parameter.description,
                vault_name=parameter.vault_name,
                username_key=parameter.username_key,
                password_key=parameter.password_key,
                totp_secret_key=parameter.totp_secret_key,
                created_at=now,
                modified_at=now,
            )
        elif parameter.parameter_type == ParameterType.BITWARDEN_LOGIN_CREDENTIAL:
            if not parameter.bitwarden_collection_id and not parameter.bitwarden_item_id:
                raise WorkflowParameterMissingRequiredValue(
                    workflow_parameter_type=ParameterType.BITWARDEN_LOGIN_CREDENTIAL,
                    workflow_parameter_key=parameter.key,
                    required_value="bitwarden_collection_id or bitwarden_item_id",
                )
            if parameter.bitwarden_collection_id and not parameter.url_parameter_key:
                raise WorkflowParameterMissingRequiredValue(
                    workflow_parameter_type=ParameterType.BITWARDEN_LOGIN_CREDENTIAL,
                    workflow_parameter_key=parameter.key,
                    required_value="url_parameter_key",
                )
            parameters[parameter.key] = BitwardenLoginCredentialParameter(
                bitwarden_login_credential_parameter_id=generate_bitwarden_login_credential_parameter_id(),
                workflow_id=workflow_id,
                bitwarden_client_id_aws_secret_key=parameter.bitwarden_client_id_aws_secret_key,
                bitwarden_client_secret_aws_secret_key=parameter.bitwarden_client_secret_aws_secret_key,
                bitwarden_master_password_aws_secret_key=parameter.bitwarden_master_password_aws_secret_key,
                url_parameter_key=parameter.url_parameter_key,
                key=parameter.key,
                description=parameter.description,
                bitwarden_collection_id=parameter.bitwarden_collection_id,
                bitwarden_item_id=parameter.bitwarden_item_id,
                created_at=now,
                modified_at=now,
            )
        elif parameter.parameter_type == ParameterType.BITWARDEN_SENSITIVE_INFORMATION:
            parameters[parameter.key] = BitwardenSensitiveInformationParameter(
                bitwarden_sensitive_information_parameter_id=generate_bitwarden_sensitive_information_parameter_id(),
                workflow_id=workflow_id,
                bitwarden_client_id_aws_secret_key=parameter.bitwarden_client_id_aws_secret_key,
                bitwarden_client_secret_aws_secret_key=parameter.bitwarden_client_secret_aws_secret_key,
                bitwarden_master_password_aws_secret_key=parameter.bitwarden_master_password_aws_secret_key,
                # TODO: remove "# type: ignore" after ensuring bitwarden_collection_id is always set
                bitwarden_collection_id=parameter.bitwarden_collection_id,  # type: ignore
                bitwarden_identity_key=parameter.bitwarden_identity_key,
                bitwarden_identity_fields=parameter.bitwarden_identity_fields,
                key=parameter.key,
                description=parameter.description,
                created_at=now,
                modified_at=now,
            )
        elif parameter.parameter_type == ParameterType.BITWARDEN_CREDIT_CARD_DATA:
            parameters[parameter.key] = BitwardenCreditCardDataParameter(
                bitwarden_credit_card_data_parameter_id=generate_bitwarden_credit_card_data_parameter_id(),
                workflow_id=workflow_id,
                bitwarden_client_id_aws_secret_key=parameter.bitwarden_client_id_aws_secret_key,
                bitwarden_client_secret_aws_secret_key=parameter.bitwarden_client_secret_aws_secret_key,
                bitwarden_master_password_aws_secret_key=parameter.bitwarden_master_password_aws_secret_key,
                # TODO: remove "# type: ignore" after ensuring bitwarden_collection_id is always set
                bitwarden_collection_id=parameter.bitwarden_collection_id,  # type: ignore
                bitwarden_item_id=parameter.bitwarden_item_id,  # type: ignore
                key=parameter.key,
                description=parameter.description,
                created_at=now,
                modified_at=now,
            )
        elif parameter.parameter_type == ParameterType.WORKFLOW:
            default_value = parameter.workflow_parameter_type.convert_value(
                json.dumps(parameter.default_value)
                if parameter.workflow_parameter_type == WorkflowParameterType.JSON
                else parameter.default_value
            )
            parameters[parameter.key] = WorkflowParameter(
                workflow_parameter_id=generate_workflow_parameter_id(),
                workflow_parameter_type=parameter.workflow_parameter_type,
                workflow_id=workflow_id,
                key=parameter.key,
                default_value=default_value,
                description=parameter.description,
                created_at=now,
                modified_at=now,
            )
        elif parameter.parameter_type == ParameterType.OUTPUT:
            parameters[parameter.key] = OutputParameter(
                output_parameter_id=generate_output_parameter_id(),
                workflow_id=workflow_id,
                key=parameter.key,
                description=parameter.description,
                created_at=now,
                modified_at=now,
            )
        elif parameter.parameter_type == ParameterType.CONTEXT:
            context_parameter_yamls.append(parameter)
        else:
            LOG.error(f"Invalid parameter type {parameter.parameter_type}")

    # Now we can process the context parameters since all other parameters have been created
    for context_parameter in context_parameter_yamls:
        if context_parameter.source_parameter_key not in parameters:
            raise ContextParameterSourceNotDefined(
                context_parameter_key=context_parameter.key,
                source_key=context_parameter.source_parameter_key,
            )

        if context_parameter.key in parameters:
            LOG.error(f"Duplicate parameter key {context_parameter.key}")
            duplicate_parameter_keys.add(context_parameter.key)
            continue

        # We're only adding the context parameter to the parameters dict, we're not creating it in the database
        # It'll only be stored in the `workflow.workflow_definition`
        # todo (kerem): should we have a database table for context parameters?
        parameters[context_parameter.key] = ContextParameter(
            key=context_parameter.key,
            description=context_parameter.description,
            source=parameters[context_parameter.source_parameter_key],
            # Context parameters don't have a default value, the value always depends on the source parameter
            value=None,
        )

    if duplicate_parameter_keys:
        raise WorkflowDefinitionHasDuplicateParameterKeys(duplicate_keys=duplicate_parameter_keys)

    # Validate that all blocks reference defined parameters
    undefined_parameters = _collect_undefined_parameters(workflow_definition_yaml.blocks, parameters)
    if undefined_parameters:
        raise WorkflowDefinitionHasUndefinedParameters(undefined_parameters=undefined_parameters)

    # Create blocks from the request
    block_label_mapping = {}
    blocks: list[BlockTypeVar] = []
    for block_yaml in workflow_definition_yaml.blocks:
        block = block_yaml_to_block(block_yaml, parameters)
        blocks.append(block)
        block_label_mapping[block.label] = block

    # Set the blocks for the workflow definition and derive DAG version metadata
    dag_version = workflow_definition_yaml.version
    if dag_version is None:
        dag_version = 2 if _has_dag_metadata(workflow_definition_yaml.blocks) else 1

    workflow_definition = WorkflowDefinition(
        parameters=parameters.values(),
        blocks=blocks,
        version=dag_version,
        finally_block_label=workflow_definition_yaml.finally_block_label,
    )

    LOG.info(
        "Created workflow from request",
        parameter_keys=[parameter.key for parameter in parameters.values()],
        block_labels=[block.label for block in blocks],
        workflow_id=workflow_id,
    )

    return workflow_definition


def _create_all_output_parameters_for_workflow(
    workflow_id: str, block_yamls: list[BLOCK_YAML_TYPES]
) -> dict[str, OutputParameter]:
    output_parameters = {}
    for block_yaml in block_yamls:
        output_parameter_key = f"{block_yaml.label}_output"
        output_parameter = OutputParameter(
            output_parameter_id=generate_output_parameter_id(),
            key=output_parameter_key,
            description=f"Output parameter for block {block_yaml.label}",
            workflow_id=workflow_id,
            created_at=datetime.utcnow(),
            modified_at=datetime.utcnow(),
        )
        output_parameters[block_yaml.label] = output_parameter
        # Recursively create output parameters for for-loop blocks
        if isinstance(block_yaml, ForLoopBlockYAML):
            output_parameters.update(
                _create_all_output_parameters_for_workflow(workflow_id=workflow_id, block_yamls=block_yaml.loop_blocks)
            )
    return output_parameters


def _build_block_kwargs(
    block_yaml: BLOCK_YAML_TYPES,
    output_parameter: OutputParameter,
) -> dict[str, Any]:
    return {
        "label": block_yaml.label,
        "next_block_label": block_yaml.next_block_label,
        "output_parameter": output_parameter,
        "continue_on_failure": block_yaml.continue_on_failure,
        "next_loop_on_failure": block_yaml.next_loop_on_failure,
        "model": block_yaml.model,
    }


def block_yaml_to_block(
    block_yaml: BLOCK_YAML_TYPES,
    parameters: dict[str, PARAMETER_TYPE],
) -> BlockTypeVar:
    output_parameter = cast(OutputParameter, parameters[f"{block_yaml.label}_output"])
    base_kwargs = _build_block_kwargs(block_yaml, output_parameter)
    if block_yaml.block_type == BlockType.TASK:
        task_block_parameters = _resolve_block_parameters(block_yaml, parameters)
        return TaskBlock(
            **base_kwargs,
            url=block_yaml.url,
            title=block_yaml.title,
            engine=block_yaml.engine,
            parameters=task_block_parameters,
            navigation_goal=block_yaml.navigation_goal,
            data_extraction_goal=block_yaml.data_extraction_goal,
            data_schema=block_yaml.data_schema,
            error_code_mapping=block_yaml.error_code_mapping,
            max_steps_per_run=block_yaml.max_steps_per_run,
            max_retries=block_yaml.max_retries,
            complete_on_download=block_yaml.complete_on_download,
            download_suffix=block_yaml.download_suffix,
            totp_verification_url=block_yaml.totp_verification_url,
            totp_identifier=block_yaml.totp_identifier,
            disable_cache=block_yaml.disable_cache,
            complete_criterion=block_yaml.complete_criterion,
            terminate_criterion=block_yaml.terminate_criterion,
            complete_verification=block_yaml.complete_verification,
            include_action_history_in_verification=block_yaml.include_action_history_in_verification,
        )
    elif block_yaml.block_type == BlockType.FOR_LOOP:
        loop_blocks = [block_yaml_to_block(loop_block, parameters) for loop_block in block_yaml.loop_blocks]

        loop_over_parameter: Parameter | None = None
        if block_yaml.loop_over_parameter_key:
            loop_over_parameter = parameters[block_yaml.loop_over_parameter_key]

        if block_yaml.loop_variable_reference:
            # it's backaward compatible with jinja style parameter and context paramter
            # we trim the format like {{ loop_key }} into loop_key to initialize the context parater,
            # otherwise it might break the context parameter initialization chain, blow up the worklofw parameters
            # TODO: consider remove this if we totally give up context parameter
            trimmed_key = block_yaml.loop_variable_reference.strip(" {}")
            if trimmed_key in parameters:
                loop_over_parameter = parameters[trimmed_key]

        if loop_over_parameter is None and not block_yaml.loop_variable_reference:
            raise InvalidWorkflowDefinition(
                f"For loop block '{block_yaml.label}' requires either loop_over_parameter_key or loop_variable_reference"
            )

        return ForLoopBlock(
            **base_kwargs,
            loop_over=loop_over_parameter,
            loop_variable_reference=block_yaml.loop_variable_reference,
            loop_blocks=loop_blocks,
            complete_if_empty=block_yaml.complete_if_empty,
            data_schema=block_yaml.data_schema,
        )
    elif block_yaml.block_type == BlockType.CONDITIONAL:
        branch_conditions = []
        for branch in block_yaml.branch_conditions:
            branch_criteria = None
            if branch.criteria:
                if branch.criteria.criteria_type == "prompt":
                    branch_criteria = PromptBranchCriteria(
                        criteria_type=branch.criteria.criteria_type,
                        expression=branch.criteria.expression,
                        description=branch.criteria.description,
                    )
                else:
                    branch_criteria = JinjaBranchCriteria(
                        criteria_type=branch.criteria.criteria_type,
                        expression=branch.criteria.expression,
                        description=branch.criteria.description,
                    )

            branch_conditions.append(
                BranchCondition(
                    criteria=branch_criteria,
                    next_block_label=branch.next_block_label,
                    description=branch.description,
                    is_default=branch.is_default,
                )
            )

        return ConditionalBlock(
            **base_kwargs,
            branch_conditions=branch_conditions,
        )
    elif block_yaml.block_type == BlockType.CODE:
        return CodeBlock(
            **base_kwargs,
            code=block_yaml.code,
            parameters=_resolve_block_parameters(block_yaml, parameters),
        )
    elif block_yaml.block_type == BlockType.TEXT_PROMPT:
        return TextPromptBlock(
            **base_kwargs,
            llm_key=block_yaml.llm_key,
            prompt=block_yaml.prompt,
            parameters=_resolve_block_parameters(block_yaml, parameters),
            json_schema=block_yaml.json_schema,
        )
    elif block_yaml.block_type == BlockType.DOWNLOAD_TO_S3:
        return DownloadToS3Block(
            **base_kwargs,
            url=block_yaml.url,
        )
    elif block_yaml.block_type == BlockType.UPLOAD_TO_S3:
        return UploadToS3Block(
            **base_kwargs,
            path=block_yaml.path,
        )
    elif block_yaml.block_type == BlockType.FILE_UPLOAD:
        return FileUploadBlock(
            **base_kwargs,
            storage_type=block_yaml.storage_type,
            s3_bucket=block_yaml.s3_bucket,
            aws_access_key_id=block_yaml.aws_access_key_id,
            aws_secret_access_key=block_yaml.aws_secret_access_key,
            region_name=block_yaml.region_name,
            azure_storage_account_name=block_yaml.azure_storage_account_name,
            azure_storage_account_key=block_yaml.azure_storage_account_key,
            azure_blob_container_name=block_yaml.azure_blob_container_name,
            path=block_yaml.path,
        )
    elif block_yaml.block_type == BlockType.SEND_EMAIL:
        return SendEmailBlock(
            **base_kwargs,
            smtp_host=parameters[block_yaml.smtp_host_secret_parameter_key],
            smtp_port=parameters[block_yaml.smtp_port_secret_parameter_key],
            smtp_username=parameters[block_yaml.smtp_username_secret_parameter_key],
            smtp_password=parameters[block_yaml.smtp_password_secret_parameter_key],
            sender=block_yaml.sender,
            recipients=block_yaml.recipients,
            subject=block_yaml.subject,
            body=block_yaml.body,
            file_attachments=block_yaml.file_attachments or [],
        )
    elif block_yaml.block_type == BlockType.FILE_URL_PARSER:
        return FileParserBlock(
            **base_kwargs,
            file_url=block_yaml.file_url,
            file_type=block_yaml.file_type,
            json_schema=block_yaml.json_schema,
        )
    elif block_yaml.block_type == BlockType.PDF_PARSER:
        return PDFParserBlock(
            **base_kwargs,
            file_url=block_yaml.file_url,
            json_schema=block_yaml.json_schema,
        )
    elif block_yaml.block_type == BlockType.VALIDATION:
        validation_block_parameters = _resolve_block_parameters(block_yaml, parameters)

        if not block_yaml.complete_criterion and not block_yaml.terminate_criterion:
            raise InvalidWorkflowDefinition(
                f"Validation block '{block_yaml.label}' requires at least one of complete_criterion or terminate_criterion"
            )

        return ValidationBlock(
            **base_kwargs,
            task_type=TaskType.validation,
            parameters=validation_block_parameters,
            complete_criterion=block_yaml.complete_criterion,
            terminate_criterion=block_yaml.terminate_criterion,
            error_code_mapping=block_yaml.error_code_mapping,
            # Should only need one step for validation block, but we allow 2 in case the LLM has an unexpected failure and we need to retry.
            max_steps_per_run=2,
        )

    elif block_yaml.block_type == BlockType.ACTION:
        action_block_parameters = _resolve_block_parameters(block_yaml, parameters)

        if not block_yaml.navigation_goal:
            raise InvalidWorkflowDefinition(f"Action block '{block_yaml.label}' requires navigation_goal")

        return ActionBlock(
            **base_kwargs,
            url=block_yaml.url,
            title=block_yaml.title,
            engine=block_yaml.engine,
            task_type=TaskType.action,
            parameters=action_block_parameters,
            navigation_goal=block_yaml.navigation_goal,
            error_code_mapping=block_yaml.error_code_mapping,
            max_retries=block_yaml.max_retries,
            complete_on_download=block_yaml.complete_on_download,
            download_suffix=block_yaml.download_suffix,
            totp_verification_url=block_yaml.totp_verification_url,
            totp_identifier=block_yaml.totp_identifier,
            disable_cache=block_yaml.disable_cache,
            # DO NOT run complete verification for action block
            complete_verification=False,
            max_steps_per_run=1,
        )

    elif block_yaml.block_type == BlockType.NAVIGATION:
        navigation_block_parameters = _resolve_block_parameters(block_yaml, parameters)
        return NavigationBlock(
            **base_kwargs,
            url=block_yaml.url,
            title=block_yaml.title,
            engine=block_yaml.engine,
            parameters=navigation_block_parameters,
            navigation_goal=block_yaml.navigation_goal,
            error_code_mapping=block_yaml.error_code_mapping,
            max_steps_per_run=block_yaml.max_steps_per_run,
            max_retries=block_yaml.max_retries,
            complete_on_download=block_yaml.complete_on_download,
            download_suffix=block_yaml.download_suffix,
            totp_verification_url=block_yaml.totp_verification_url,
            totp_identifier=block_yaml.totp_identifier,
            disable_cache=block_yaml.disable_cache,
            complete_criterion=block_yaml.complete_criterion,
            terminate_criterion=block_yaml.terminate_criterion,
            complete_verification=block_yaml.complete_verification,
            include_action_history_in_verification=block_yaml.include_action_history_in_verification,
        )

    elif block_yaml.block_type == BlockType.HUMAN_INTERACTION:
        return HumanInteractionBlock(
            **base_kwargs,
            instructions=block_yaml.instructions,
            positive_descriptor=block_yaml.positive_descriptor,
            negative_descriptor=block_yaml.negative_descriptor,
            timeout_seconds=block_yaml.timeout_seconds,
            # --
            sender=block_yaml.sender,
            recipients=block_yaml.recipients,
            subject=block_yaml.subject,
            body=block_yaml.body,
        )

    elif block_yaml.block_type == BlockType.EXTRACTION:
        extraction_block_parameters = _resolve_block_parameters(block_yaml, parameters)
        return ExtractionBlock(
            **base_kwargs,
            url=block_yaml.url,
            title=block_yaml.title,
            engine=block_yaml.engine,
            parameters=extraction_block_parameters,
            data_extraction_goal=block_yaml.data_extraction_goal,
            data_schema=block_yaml.data_schema,
            max_steps_per_run=block_yaml.max_steps_per_run,
            max_retries=block_yaml.max_retries,
            disable_cache=block_yaml.disable_cache,
            complete_verification=False,
        )

    elif block_yaml.block_type == BlockType.LOGIN:
        login_block_parameters = _resolve_block_parameters(block_yaml, parameters)
        return LoginBlock(
            **base_kwargs,
            url=block_yaml.url,
            title=block_yaml.title,
            engine=block_yaml.engine,
            parameters=login_block_parameters,
            navigation_goal=block_yaml.navigation_goal,
            error_code_mapping=block_yaml.error_code_mapping,
            max_steps_per_run=block_yaml.max_steps_per_run,
            max_retries=block_yaml.max_retries,
            totp_verification_url=block_yaml.totp_verification_url,
            totp_identifier=block_yaml.totp_identifier,
            disable_cache=block_yaml.disable_cache,
            complete_criterion=block_yaml.complete_criterion,
            terminate_criterion=block_yaml.terminate_criterion,
            complete_verification=block_yaml.complete_verification,
        )

    elif block_yaml.block_type == BlockType.WAIT:
        if block_yaml.wait_sec <= 0 or block_yaml.wait_sec > settings.WORKFLOW_WAIT_BLOCK_MAX_SEC:
            raise InvalidWaitBlockTime(settings.WORKFLOW_WAIT_BLOCK_MAX_SEC)

        return WaitBlock(
            **base_kwargs,
            wait_sec=block_yaml.wait_sec,
        )

    elif block_yaml.block_type == BlockType.FILE_DOWNLOAD:
        file_download_block_parameters = _resolve_block_parameters(block_yaml, parameters)
        return FileDownloadBlock(
            **base_kwargs,
            url=block_yaml.url,
            title=block_yaml.title,
            engine=block_yaml.engine,
            parameters=file_download_block_parameters,
            navigation_goal=block_yaml.navigation_goal,
            error_code_mapping=block_yaml.error_code_mapping,
            max_steps_per_run=block_yaml.max_steps_per_run,
            max_retries=block_yaml.max_retries,
            download_suffix=block_yaml.download_suffix,
            totp_verification_url=block_yaml.totp_verification_url,
            totp_identifier=block_yaml.totp_identifier,
            disable_cache=block_yaml.disable_cache,
            complete_on_download=True,
            complete_verification=True,
            include_action_history_in_verification=True,
            download_timeout=block_yaml.download_timeout,
        )
    elif block_yaml.block_type == BlockType.TaskV2:
        return TaskV2Block(
            **base_kwargs,
            prompt=block_yaml.prompt,
            url=block_yaml.url,
            totp_verification_url=block_yaml.totp_verification_url,
            totp_identifier=block_yaml.totp_identifier,
            max_iterations=block_yaml.max_iterations,
            max_steps=block_yaml.max_steps,
        )
    elif block_yaml.block_type == BlockType.HTTP_REQUEST:
        http_request_block_parameters = _resolve_block_parameters(block_yaml, parameters)
        return HttpRequestBlock(
            **base_kwargs,
            method=block_yaml.method,
            url=block_yaml.url,
            headers=block_yaml.headers,
            body=block_yaml.body,
            files=block_yaml.files,
            timeout=block_yaml.timeout,
            follow_redirects=block_yaml.follow_redirects,
            download_filename=block_yaml.download_filename,
            save_response_as_file=block_yaml.save_response_as_file,
            parameters=http_request_block_parameters,
        )
    elif block_yaml.block_type == BlockType.GOTO_URL:
        return UrlBlock(
            **base_kwargs,
            url=block_yaml.url,
            complete_verification=False,
        )
    elif block_yaml.block_type == BlockType.PRINT_PAGE:
        print_page_block_parameters = _resolve_block_parameters(block_yaml, parameters)
        return PrintPageBlock(
            **base_kwargs,
            include_timestamp=block_yaml.include_timestamp,
            custom_filename=block_yaml.custom_filename,
            format=block_yaml.format,
            landscape=block_yaml.landscape,
            print_background=block_yaml.print_background,
            parameters=print_page_block_parameters,
        )

    raise ValueError(f"Invalid block type {block_yaml.block_type}")


def _collect_undefined_parameters(
    block_yamls: list[BLOCK_YAML_TYPES],
    parameters: dict[str, PARAMETER_TYPE],
) -> dict[str, list[str]]:
    """
    Collect all undefined parameters referenced by blocks (including nested blocks in for_loop).
    Returns a dict mapping block labels to lists of undefined parameter keys.
    """
    undefined_params: dict[str, list[str]] = {}

    for block_yaml in block_yamls:
        if hasattr(block_yaml, "parameter_keys") and block_yaml.parameter_keys:
            undefined_for_block = [param_key for param_key in block_yaml.parameter_keys if param_key not in parameters]
            if undefined_for_block:
                undefined_params[block_yaml.label] = undefined_for_block

        # Recursively check nested blocks in for_loop
        if isinstance(block_yaml, ForLoopBlockYAML) and block_yaml.loop_blocks:
            nested_undefined = _collect_undefined_parameters(block_yaml.loop_blocks, parameters)
            undefined_params.update(nested_undefined)

    return undefined_params


def _resolve_block_parameters(
    block_yaml: BLOCK_YAML_TYPES,
    parameters: dict[str, PARAMETER_TYPE],
) -> list[PARAMETER_TYPE]:
    parameter_keys = getattr(block_yaml, "parameter_keys", None)
    return [parameters[parameter_key] for parameter_key in parameter_keys] if parameter_keys else []


def _has_dag_metadata(block_yamls: list[BLOCK_YAML_TYPES]) -> bool:
    for block_yaml in block_yamls:
        if block_yaml.next_block_label:
            return True
        if isinstance(block_yaml, ForLoopBlockYAML) and _has_dag_metadata(block_yaml.loop_blocks):
            return True
    return False



# This file was auto-generated by Fern from our API Definition.

import typing

from .credit_card_credential_response import CreditCardCredentialResponse
from .password_credential_response import PasswordCredentialResponse
from .secret_credential_response import SecretCredentialResponse

CredentialResponseCredential = typing.Union[
    PasswordCredentialResponse, CreditCardCredentialResponse, SecretCredentialResponse
]



from __future__ import annotations

import json
from dataclasses import asdict, dataclass
from datetime import datetime, timezone
from pathlib import Path

STATE_DIR = Path.home() / ".skyvern"
STATE_FILE = STATE_DIR / "state.json"

_TTL_SECONDS = 86400  # 24 hours


@dataclass
class CLIState:
    session_id: str | None = None
    cdp_url: str | None = None
    mode: str | None = None  # "cloud", "local", or "cdp"
    created_at: str | None = None


def save_state(state: CLIState) -> None:
    STATE_DIR.mkdir(parents=True, exist_ok=True)
    STATE_DIR.chmod(0o700)
    data = asdict(state)
    data["created_at"] = datetime.now(timezone.utc).isoformat()
    STATE_FILE.write_text(json.dumps(data))
    STATE_FILE.chmod(0o600)


def load_state() -> CLIState | None:
    if not STATE_FILE.exists():
        return None
    try:
        data = json.loads(STATE_FILE.read_text())
        created_at = data.get("created_at")
        if created_at:
            age = (datetime.now(timezone.utc) - datetime.fromisoformat(created_at)).total_seconds()
            if age > _TTL_SECONDS:
                return None
        return CLIState(**{k: v for k, v in data.items() if k in CLIState.__dataclass_fields__})
    except Exception:
        return None


def clear_state() -> None:
    if STATE_FILE.exists():
        STATE_FILE.unlink()



import base64
import hashlib

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

from skyvern.forge.sdk.encrypt.base import BaseEncryptor, EncryptMethod

default_iv = hashlib.md5(b"deterministic_iv_0123456789").digest()
default_salt = hashlib.md5(b"deterministic_salt_0123456789").digest()


class AES(BaseEncryptor):
    def __init__(self, *, secret_key: str, salt: str | None = None, iv: str | None = None) -> None:
        self.secret_key = hashlib.md5(secret_key.encode("utf-8")).digest()
        self.salt = hashlib.md5(salt.encode("utf-8")).digest() if salt else default_salt
        self.iv = hashlib.md5(iv.encode("utf-8")).digest() if iv else default_iv

    def method(self) -> EncryptMethod:
        return EncryptMethod.AES

    def _derive_key(self) -> bytes:
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=self.salt,
            iterations=100000,
        )
        return kdf.derive(self.secret_key)

    async def encrypt(self, plaintext: str) -> str:
        try:
            key = self._derive_key()
            cipher = Cipher(algorithms.AES(key), modes.CBC(self.iv))
            encryptor = cipher.encryptor()
            padded_plaintext = self._pad(plaintext.encode("utf-8"))
            ciphertext = encryptor.update(padded_plaintext) + encryptor.finalize()
            return base64.b64encode(ciphertext).decode("utf-8")
        except Exception as e:
            raise Exception("Failed to encrypt token") from e

    async def decrypt(self, ciphertext: str) -> str:
        try:
            encrypted_data = base64.b64decode(ciphertext.encode("utf-8"))
            key = self._derive_key()
            cipher = Cipher(algorithms.AES(key), modes.CBC(self.iv))
            decryptor = cipher.decryptor()
            padded_plaintext = decryptor.update(encrypted_data) + decryptor.finalize()
            plaintext = self._unpad(padded_plaintext)
            return plaintext.decode("utf-8")
        except Exception as e:
            raise Exception("Failed to decrypt token") from e

    def _pad(self, data: bytes) -> bytes:
        block_size = 16
        padding_length = block_size - (len(data) % block_size)
        padding = bytes([padding_length] * padding_length)
        return data + padding

    def _unpad(self, data: bytes) -> bytes:
        padding_length = data[-1]
        return data[:-padding_length]



"""Credential CLI commands with MCP-parity output and validation."""

from __future__ import annotations

import asyncio
import json
import sys
from typing import Any, Callable, Coroutine

import typer
from dotenv import load_dotenv

from skyvern.config import settings
from skyvern.utils.env_paths import resolve_backend_env_path

from .commands._output import output, output_error
from .mcp_tools.credential import skyvern_credential_delete as tool_credential_delete
from .mcp_tools.credential import skyvern_credential_get as tool_credential_get
from .mcp_tools.credential import skyvern_credential_list as tool_credential_list

credential_app = typer.Typer(
    help="MCP-parity credential commands (list/get/delete). Use `skyvern credentials add` for secure creation.",
    no_args_is_help=True,
)


def _emit_tool_result(result: dict[str, Any], *, json_output: bool) -> None:
    if json_output:
        json.dump(result, sys.stdout, indent=2, default=str)
        sys.stdout.write("\n")
        if not result.get("ok", False):
            raise SystemExit(1)
        return

    if result.get("ok", False):
        output(result.get("data"), action=str(result.get("action", "")), json_mode=False)
        return

    err = result.get("error") or {}
    output_error(str(err.get("message") or "Unknown error"), hint=str(err.get("hint") or ""), json_mode=False)


def _run_tool(
    runner: Callable[[], Coroutine[Any, Any, dict[str, Any]]],
    *,
    json_output: bool,
    hint_on_exception: str,
) -> None:
    try:
        result: dict[str, Any] = asyncio.run(runner())
        _emit_tool_result(result, json_output=json_output)
    except typer.BadParameter:
        raise
    except Exception as e:
        output_error(str(e), hint=hint_on_exception, json_mode=json_output)


@credential_app.callback()
def credential_callback(
    api_key: str | None = typer.Option(
        None,
        "--api-key",
        help="Skyvern API key",
        envvar="SKYVERN_API_KEY",
    ),
) -> None:
    """Load environment and optional API key override."""
    load_dotenv(resolve_backend_env_path())
    if api_key:
        settings.SKYVERN_API_KEY = api_key


@credential_app.command("list")
def credential_list(
    page: int = typer.Option(1, "--page", min=1, help="Page number (1-based)."),
    page_size: int = typer.Option(10, "--page-size", min=1, max=100, help="Results per page."),
    json_output: bool = typer.Option(False, "--json", help="Output as JSON."),
) -> None:
    """List stored credentials (metadata only)."""

    async def _run() -> dict[str, Any]:
        return await tool_credential_list(page=page, page_size=page_size)

    _run_tool(_run, json_output=json_output, hint_on_exception="Check your API key and Skyvern connection.")


@credential_app.command("get")
def credential_get(
    credential_id: str = typer.Option(..., "--id", "--credential-id", help="Credential ID (starts with cred_)."),
    json_output: bool = typer.Option(False, "--json", help="Output as JSON."),
) -> None:
    """Get credential metadata by ID."""

    async def _run() -> dict[str, Any]:
        return await tool_credential_get(credential_id=credential_id)

    _run_tool(_run, json_output=json_output, hint_on_exception="Check your API key and credential ID.")


@credential_app.command("delete")
def credential_delete(
    credential_id: str = typer.Option(..., "--id", "--credential-id", help="Credential ID (starts with cred_)."),
    json_output: bool = typer.Option(False, "--json", help="Output as JSON."),
) -> None:
    """Delete a credential by ID."""

    async def _run() -> dict[str, Any]:
        return await tool_credential_delete(credential_id=credential_id)

    _run_tool(_run, json_output=json_output, hint_on_exception="Check your API key and credential ID.")



"""Workflow-related CLI commands with MCP-parity flags and output."""

from __future__ import annotations

import asyncio
import json
import sys
from pathlib import Path
from typing import Any, Callable, Coroutine

import typer
from dotenv import load_dotenv

from skyvern.config import settings
from skyvern.utils.env_paths import resolve_backend_env_path

from .commands._output import output, output_error
from .mcp_tools.workflow import skyvern_workflow_cancel as tool_workflow_cancel
from .mcp_tools.workflow import skyvern_workflow_create as tool_workflow_create
from .mcp_tools.workflow import skyvern_workflow_delete as tool_workflow_delete
from .mcp_tools.workflow import skyvern_workflow_get as tool_workflow_get
from .mcp_tools.workflow import skyvern_workflow_list as tool_workflow_list
from .mcp_tools.workflow import skyvern_workflow_run as tool_workflow_run
from .mcp_tools.workflow import skyvern_workflow_status as tool_workflow_status
from .mcp_tools.workflow import skyvern_workflow_update as tool_workflow_update

workflow_app = typer.Typer(help="Manage Skyvern workflows.", no_args_is_help=True)


def _emit_tool_result(result: dict[str, Any], *, json_output: bool) -> None:
    if json_output:
        json.dump(result, sys.stdout, indent=2, default=str)
        sys.stdout.write("\n")
        if not result.get("ok", False):
            raise SystemExit(1)
        return

    if result.get("ok", False):
        output(result.get("data"), action=str(result.get("action", "")), json_mode=False)
        return

    err = result.get("error") or {}
    output_error(str(err.get("message", "Unknown error")), hint=str(err.get("hint", "")), json_mode=False)


def _run_tool(
    runner: Callable[[], Coroutine[Any, Any, dict[str, Any]]],
    *,
    json_output: bool,
    hint_on_exception: str,
) -> None:
    try:
        result: dict[str, Any] = asyncio.run(runner())
        _emit_tool_result(result, json_output=json_output)
    except typer.BadParameter:
        raise
    except Exception as e:
        output_error(str(e), hint=hint_on_exception, json_mode=json_output)


def _resolve_inline_or_file(value: str | None, *, param_name: str) -> str | None:
    if value is None or not value.startswith("@"):
        return value

    file_path = value[1:]
    if not file_path:
        raise typer.BadParameter(f"{param_name} file path cannot be empty after '@'.")

    path = Path(file_path).expanduser()
    try:
        return path.read_text(encoding="utf-8")
    except OSError as e:
        raise typer.BadParameter(f"Unable to read {param_name} file '{path}': {e}") from e


@workflow_app.callback()
def workflow_callback(
    api_key: str | None = typer.Option(
        None,
        "--api-key",
        help="Skyvern API key",
        envvar="SKYVERN_API_KEY",
    ),
) -> None:
    """Load workflow CLI environment and optional API key override."""
    load_dotenv(resolve_backend_env_path())
    if api_key:
        settings.SKYVERN_API_KEY = api_key


@workflow_app.command("list")
def workflow_list(
    search: str | None = typer.Option(
        None,
        "--search",
        help="Search across workflow titles, folder names, and parameter metadata.",
    ),
    page: int = typer.Option(1, "--page", min=1, help="Page number (1-based)."),
    page_size: int = typer.Option(10, "--page-size", min=1, max=100, help="Results per page."),
    only_workflows: bool = typer.Option(
        False,
        "--only-workflows",
        help="Only return multi-step workflows (exclude saved tasks).",
    ),
    json_output: bool = typer.Option(False, "--json", help="Output as JSON."),
) -> None:
    """List workflows."""

    async def _run() -> dict[str, Any]:
        return await tool_workflow_list(
            search=search,
            page=page,
            page_size=page_size,
            only_workflows=only_workflows,
        )

    _run_tool(_run, json_output=json_output, hint_on_exception="Check your API key and workflow list filters.")


@workflow_app.command("get")
def workflow_get(
    workflow_id: str = typer.Option(..., "--id", help="Workflow permanent ID (wpid_...)."),
    version: int | None = typer.Option(None, "--version", min=1, help="Specific version to retrieve."),
    json_output: bool = typer.Option(False, "--json", help="Output as JSON."),
) -> None:
    """Get a workflow definition by ID."""

    async def _run() -> dict[str, Any]:
        return await tool_workflow_get(workflow_id=workflow_id, version=version)

    _run_tool(_run, json_output=json_output, hint_on_exception="Check your API key and workflow ID.")


@workflow_app.command("create")
def workflow_create(
    definition: str = typer.Option(
        ...,
        "--definition",
        help="Workflow definition as YAML/JSON string or @file path.",
    ),
    definition_format: str = typer.Option(
        "auto",
        "--format",
        help="Definition format: json, yaml, or auto.",
    ),
    folder_id: str | None = typer.Option(None, "--folder-id", help="Folder ID (fld_...) for the workflow."),
    json_output: bool = typer.Option(False, "--json", help="Output as JSON."),
) -> None:
    """Create a workflow."""

    async def _run() -> dict[str, Any]:
        resolved_definition = _resolve_inline_or_file(definition, param_name="definition")
        assert resolved_definition is not None
        return await tool_workflow_create(
            definition=resolved_definition,
            format=definition_format,
            folder_id=folder_id,
        )

    _run_tool(_run, json_output=json_output, hint_on_exception="Check the workflow definition syntax.")


@workflow_app.command("update")
def workflow_update(
    workflow_id: str = typer.Option(..., "--id", help="Workflow permanent ID (wpid_...)."),
    definition: str = typer.Option(
        ...,
        "--definition",
        help="Updated workflow definition as YAML/JSON string or @file path.",
    ),
    definition_format: str = typer.Option(
        "auto",
        "--format",
        help="Definition format: json, yaml, or auto.",
    ),
    json_output: bool = typer.Option(False, "--json", help="Output as JSON."),
) -> None:
    """Update a workflow definition."""

    async def _run() -> dict[str, Any]:
        resolved_definition = _resolve_inline_or_file(definition, param_name="definition")
        assert resolved_definition is not None
        return await tool_workflow_update(
            workflow_id=workflow_id,
            definition=resolved_definition,
            format=definition_format,
        )

    _run_tool(_run, json_output=json_output, hint_on_exception="Check the workflow ID and definition syntax.")


@workflow_app.command("delete")
def workflow_delete(
    workflow_id: str = typer.Option(..., "--id", help="Workflow permanent ID (wpid_...)."),
    force: bool = typer.Option(False, "--force", help="Confirm permanent deletion."),
    json_output: bool = typer.Option(False, "--json", help="Output as JSON."),
) -> None:
    """Delete a workflow."""

    async def _run() -> dict[str, Any]:
        return await tool_workflow_delete(workflow_id=workflow_id, force=force)

    _run_tool(_run, json_output=json_output, hint_on_exception="Check the workflow ID and your permissions.")


@workflow_app.command("run")
def workflow_run(
    workflow_id: str = typer.Option(..., "--id", help="Workflow permanent ID (wpid_...)."),
    params: str | None = typer.Option(
        None,
        "--params",
        "--parameters",
        "-p",
        help="Workflow parameters as JSON string or @file path.",
    ),
    session: str | None = typer.Option(None, "--session", help="Browser session ID (pbs_...) to reuse."),
    webhook: str | None = typer.Option(None, "--webhook", help="Status webhook callback URL."),
    proxy: str | None = typer.Option(None, "--proxy", help="Proxy location (e.g., RESIDENTIAL)."),
    wait: bool = typer.Option(False, "--wait", help="Wait for workflow completion before returning."),
    timeout: int = typer.Option(
        300,
        "--timeout",
        min=10,
        max=3600,
        help="Max wait time in seconds when --wait is set.",
    ),
    json_output: bool = typer.Option(False, "--json", help="Output as JSON."),
) -> None:
    """Run a workflow."""

    async def _run() -> dict[str, Any]:
        resolved_params = _resolve_inline_or_file(params, param_name="params")
        return await tool_workflow_run(
            workflow_id=workflow_id,
            parameters=resolved_params,
            browser_session_id=session,
            webhook_url=webhook,
            proxy_location=proxy,
            wait=wait,
            timeout_seconds=timeout,
        )

    _run_tool(_run, json_output=json_output, hint_on_exception="Check the workflow ID and run parameters.")


@workflow_app.command("status")
def workflow_status(
    run_id: str = typer.Option(..., "--run-id", help="Run ID (wr_... or tsk_v2_...)."),
    json_output: bool = typer.Option(False, "--json", help="Output as JSON."),
) -> None:
    """Get workflow run status."""

    async def _run() -> dict[str, Any]:
        return await tool_workflow_status(run_id=run_id)

    _run_tool(_run, json_output=json_output, hint_on_exception="Check the run ID and API key.")


@workflow_app.command("cancel")
def workflow_cancel(
    run_id: str = typer.Option(..., "--run-id", help="Run ID (wr_... or tsk_v2_...)."),
    json_output: bool = typer.Option(False, "--json", help="Output as JSON."),
) -> None:
    """Cancel a workflow run."""

    async def _run() -> dict[str, Any]:
        return await tool_workflow_cancel(run_id=run_id)

    _run_tool(_run, json_output=json_output, hint_on_exception="Check the run ID and API key.")



from abc import ABC, abstractmethod
from typing import BinaryIO

from skyvern.forge.sdk.artifact.models import Artifact, ArtifactType, LogEntityType
from skyvern.forge.sdk.models import Step
from skyvern.forge.sdk.schemas.ai_suggestions import AISuggestion
from skyvern.forge.sdk.schemas.files import FileInfo
from skyvern.forge.sdk.schemas.task_v2 import TaskV2, Thought
from skyvern.forge.sdk.schemas.workflow_runs import WorkflowRunBlock

# TODO: This should be a part of the ArtifactType model
FILE_EXTENTSION_MAP: dict[ArtifactType, str] = {
    ArtifactType.RECORDING: "webm",
    ArtifactType.BROWSER_CONSOLE_LOG: "log",
    ArtifactType.SCREENSHOT_LLM: "png",
    ArtifactType.SCREENSHOT_ACTION: "png",
    ArtifactType.SCREENSHOT_FINAL: "png",
    ArtifactType.SKYVERN_LOG: "log",
    ArtifactType.SKYVERN_LOG_RAW: "json",
    ArtifactType.LLM_PROMPT: "txt",
    ArtifactType.LLM_REQUEST: "json",
    ArtifactType.LLM_RESPONSE: "json",
    ArtifactType.LLM_RESPONSE_PARSED: "json",
    ArtifactType.LLM_RESPONSE_RENDERED: "json",
    ArtifactType.VISIBLE_ELEMENTS_ID_CSS_MAP: "json",
    ArtifactType.VISIBLE_ELEMENTS_ID_FRAME_MAP: "json",
    ArtifactType.VISIBLE_ELEMENTS_TREE: "json",
    ArtifactType.VISIBLE_ELEMENTS_TREE_TRIMMED: "json",
    ArtifactType.VISIBLE_ELEMENTS_TREE_IN_PROMPT: "txt",
    ArtifactType.HTML_SCRAPE: "html",
    ArtifactType.HTML_ACTION: "html",
    ArtifactType.TRACE: "zip",
    ArtifactType.HAR: "har",
    ArtifactType.HASHED_HREF_MAP: "json",
    # DEPRECATED: we're using CSS selector map now
    ArtifactType.VISIBLE_ELEMENTS_ID_XPATH_MAP: "json",
    ArtifactType.PDF: "pdf",
}


class BaseStorage(ABC):
    @abstractmethod
    def build_uri(self, *, organization_id: str, artifact_id: str, step: Step, artifact_type: ArtifactType) -> str:
        pass

    @abstractmethod
    async def retrieve_global_workflows(self) -> list[str]:
        pass

    @abstractmethod
    def build_log_uri(
        self, *, organization_id: str, log_entity_type: LogEntityType, log_entity_id: str, artifact_type: ArtifactType
    ) -> str:
        pass

    @abstractmethod
    def build_thought_uri(
        self, *, organization_id: str, artifact_id: str, thought: Thought, artifact_type: ArtifactType
    ) -> str:
        pass

    @abstractmethod
    def build_task_v2_uri(
        self, *, organization_id: str, artifact_id: str, task_v2: TaskV2, artifact_type: ArtifactType
    ) -> str:
        pass

    @abstractmethod
    def build_workflow_run_block_uri(
        self,
        *,
        organization_id: str,
        artifact_id: str,
        workflow_run_block: WorkflowRunBlock,
        artifact_type: ArtifactType,
    ) -> str:
        pass

    @abstractmethod
    def build_ai_suggestion_uri(
        self, *, organization_id: str, artifact_id: str, ai_suggestion: AISuggestion, artifact_type: ArtifactType
    ) -> str:
        pass

    @abstractmethod
    def build_script_file_uri(
        self, *, organization_id: str, script_id: str, script_version: int, file_path: str
    ) -> str:
        pass

    @abstractmethod
    async def store_artifact(self, artifact: Artifact, data: bytes) -> None:
        pass

    @abstractmethod
    async def retrieve_artifact(self, artifact: Artifact) -> bytes | None:
        pass

    @abstractmethod
    async def get_share_link(self, artifact: Artifact) -> str | None:
        pass

    @abstractmethod
    async def get_share_links(self, artifacts: list[Artifact]) -> list[str] | None:
        pass

    @abstractmethod
    async def store_artifact_from_path(self, artifact: Artifact, path: str) -> None:
        pass

    @abstractmethod
    async def save_streaming_file(self, organization_id: str, file_name: str) -> None:
        pass

    @abstractmethod
    async def get_streaming_file(self, organization_id: str, file_name: str) -> bytes | None:
        pass

    @abstractmethod
    async def store_browser_session(self, organization_id: str, workflow_permanent_id: str, directory: str) -> None:
        pass

    @abstractmethod
    async def retrieve_browser_session(self, organization_id: str, workflow_permanent_id: str) -> str | None:
        pass

    @abstractmethod
    async def store_browser_profile(self, organization_id: str, profile_id: str, directory: str) -> None:
        """Store a browser profile from a directory."""

    @abstractmethod
    async def retrieve_browser_profile(self, organization_id: str, profile_id: str) -> str | None:
        """Retrieve a browser profile to a temporary directory."""

    @abstractmethod
    async def list_downloaded_files_in_browser_session(
        self, organization_id: str, browser_session_id: str
    ) -> list[str]:
        pass

    @abstractmethod
    async def list_downloading_files_in_browser_session(
        self, organization_id: str, browser_session_id: str
    ) -> list[str]:
        pass

    @abstractmethod
    async def get_shared_downloaded_files_in_browser_session(
        self, organization_id: str, browser_session_id: str
    ) -> list[FileInfo]:
        pass

    @abstractmethod
    async def list_recordings_in_browser_session(self, organization_id: str, browser_session_id: str) -> list[str]:
        pass

    @abstractmethod
    async def get_shared_recordings_in_browser_session(
        self, organization_id: str, browser_session_id: str
    ) -> list[FileInfo]:
        pass

    @abstractmethod
    async def save_downloaded_files(self, organization_id: str, run_id: str | None) -> None:
        pass

    @abstractmethod
    async def get_downloaded_files(self, organization_id: str, run_id: str | None) -> list[FileInfo]:
        pass

    @abstractmethod
    async def save_legacy_file(
        self, *, organization_id: str, filename: str, fileObj: BinaryIO
    ) -> tuple[str, str] | None:
        pass

    @abstractmethod
    async def sync_browser_session_file(
        self,
        organization_id: str,
        browser_session_id: str,
        artifact_type: str,
        local_file_path: str,
        remote_path: str,
        date: str | None = None,
    ) -> str:
        pass

    @abstractmethod
    async def delete_browser_session_file(
        self,
        organization_id: str,
        browser_session_id: str,
        artifact_type: str,
        remote_path: str,
        date: str | None = None,
    ) -> None:
        pass

    @abstractmethod
    async def browser_session_file_exists(
        self,
        organization_id: str,
        browser_session_id: str,
        artifact_type: str,
        remote_path: str,
        date: str | None = None,
    ) -> bool:
        pass

    @abstractmethod
    async def download_uploaded_file(self, uri: str) -> bytes | None:
        pass

    @abstractmethod
    async def file_exists(self, uri: str) -> bool:
        pass

    @property
    @abstractmethod
    def storage_type(self) -> str:
        pass



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .act_action_data import ActActionData


class ActAction(UniversalBaseModel):
    """
    AI act action parameters.
    """

    intention: typing.Optional[str] = pydantic.Field(default=None)
    """
    Natural language prompt for the action
    """

    data: typing.Optional[ActActionData] = pydantic.Field(default=None)
    """
    Additional context data
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import ipaddress
from enum import Enum
from typing import Any, NamedTuple

import structlog
from fastapi import APIRouter, HTTPException, Request, status

from skyvern.config import settings
from skyvern.forge import app
from skyvern.forge.sdk.services.local_org_auth_token_service import fingerprint_token, regenerate_local_api_key
from skyvern.forge.sdk.services.org_auth_service import resolve_org_from_api_key

router = APIRouter(prefix="/internal/auth", tags=["internal"])
LOG = structlog.get_logger()


class AuthStatus(str, Enum):
    missing_env = "missing_env"
    invalid_format = "invalid_format"
    invalid = "invalid"
    expired = "expired"
    not_found = "not_found"
    ok = "ok"


class DiagnosticsResult(NamedTuple):
    status: AuthStatus
    detail: str | None
    validation: Any | None
    token: str | None


def _is_local_request(request: Request) -> bool:
    host = request.client.host if request.client else None
    if not host:
        LOG.warning("No client host found in request", client=request.client)
        return False
    try:
        addr = ipaddress.ip_address(host)
    except ValueError:
        LOG.warning("Invalid IP address in request", host=host)
        return False
    # Check if request is from Docker host (gateway IP)
    # Docker typically uses 172.x.x.x or 192.168.x.x for bridge networks
    is_local = addr.is_loopback or addr.is_private
    LOG.info(
        "Checking if request is local",
        host=host,
        is_loopback=addr.is_loopback,
        is_private=addr.is_private,
        is_local=is_local,
    )
    return is_local


def _require_local_access(request: Request) -> None:
    if settings.ENV != "local":
        raise HTTPException(status.HTTP_403_FORBIDDEN, "Endpoint only available in local env")
    if not _is_local_request(request):
        raise HTTPException(status.HTTP_403_FORBIDDEN, "Endpoint requires localhost access")


async def _evaluate_local_api_key(token: str) -> DiagnosticsResult:
    token_candidate = token.strip()
    if not token_candidate or token_candidate == "YOUR_API_KEY":
        return DiagnosticsResult(status=AuthStatus.missing_env, detail=None, validation=None, token=None)

    try:
        validation = await resolve_org_from_api_key(token_candidate, app.DATABASE)
    except HTTPException as exc:
        if exc.status_code == status.HTTP_404_NOT_FOUND:
            return DiagnosticsResult(status=AuthStatus.not_found, detail=None, token=None, validation=None)

        detail_text = exc.detail if isinstance(exc.detail, str) else None
        if exc.status_code == status.HTTP_403_FORBIDDEN:
            status_value = AuthStatus.invalid
            if detail_text and "expired" in detail_text.lower():
                status_value = AuthStatus.expired
            elif detail_text and "validate" in detail_text.lower():
                status_value = AuthStatus.invalid_format
            return DiagnosticsResult(status=status_value, detail=detail_text, token=None, validation=None)

        LOG.error("Unexpected error while diagnosing API key", status_code=exc.status_code, detail=detail_text)
        raise
    except Exception:
        LOG.error("Unexpected exception while diagnosing API key", exc_info=True)
        raise HTTPException(status.HTTP_500_INTERNAL_SERVER_ERROR, "Unable to diagnose API key")

    return DiagnosticsResult(status=AuthStatus.ok, detail=None, validation=validation, token=token_candidate)


def _emit_diagnostics(result: DiagnosticsResult) -> dict[str, object]:
    status_value = result.status.value

    if result.status is AuthStatus.ok and result.validation and result.token:
        fingerprint = fingerprint_token(result.token)
        LOG.info(
            "Local auth diagnostics",
            status=status_value,
            organization_id=result.validation.organization.organization_id,
            fingerprint=fingerprint,
            expires_at=result.validation.payload.exp,
        )
        return {
            "status": status_value,
            "organization_id": result.validation.organization.organization_id,
            "fingerprint": fingerprint,
            "expires_at": result.validation.payload.exp,
        }

    log_kwargs: dict[str, object] = {"status": status_value}
    if result.detail:
        log_kwargs["detail"] = result.detail

    LOG.warning("Local auth diagnostics", **log_kwargs)

    return {"status": status_value}


@router.post("/repair", include_in_schema=False)
async def repair_api_key(request: Request) -> dict[str, object]:
    _require_local_access(request)

    token, organization_id, backend_env_path, frontend_env_path = await regenerate_local_api_key()

    response: dict[str, object] = {
        "status": AuthStatus.ok.value,
        "organization_id": organization_id,
        "fingerprint": fingerprint_token(token),
        "api_key": token,
        "backend_env_path": backend_env_path,
    }

    if frontend_env_path:
        response["frontend_env_path"] = frontend_env_path

    return response


@router.get("/status", include_in_schema=False)
async def auth_status(request: Request) -> dict[str, object]:
    _require_local_access(request)
    token_candidate = request.headers.get("x-api-key") or ""
    result = await _evaluate_local_api_key(token_candidate)
    return _emit_diagnostics(result)



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .extensions import Extensions
from .file_info import FileInfo
from .persistent_browser_type import PersistentBrowserType


class BrowserSessionResponse(UniversalBaseModel):
    """
    Response model for browser session information.
    """

    browser_session_id: str = pydantic.Field()
    """
    Unique identifier for the browser session. browser_session_id starts with `pbs_`.
    """

    organization_id: str = pydantic.Field()
    """
    ID of the organization that owns this session
    """

    status: typing.Optional[str] = pydantic.Field(default=None)
    """
    Current status of the browser session
    """

    runnable_type: typing.Optional[str] = pydantic.Field(default=None)
    """
    Type of the current runnable associated with this session (workflow, task etc)
    """

    runnable_id: typing.Optional[str] = pydantic.Field(default=None)
    """
    ID of the current runnable
    """

    timeout: typing.Optional[int] = pydantic.Field(default=None)
    """
    Timeout in minutes for the session. Timeout is applied after the session is started. Defaults to 60 minutes.
    """

    browser_address: typing.Optional[str] = pydantic.Field(default=None)
    """
    Url for connecting to the browser
    """

    app_url: typing.Optional[str] = pydantic.Field(default=None)
    """
    Url for the browser session page
    """

    extensions: typing.Optional[typing.List[Extensions]] = pydantic.Field(default=None)
    """
    A list of extensions installed in the browser session.
    """

    browser_type: typing.Optional[PersistentBrowserType] = pydantic.Field(default=None)
    """
    The type of browser used for the session.
    """

    vnc_streaming_supported: typing.Optional[bool] = pydantic.Field(default=None)
    """
    Whether the browser session supports VNC streaming
    """

    download_path: typing.Optional[str] = pydantic.Field(default=None)
    """
    The path where the browser session downloads files
    """

    downloaded_files: typing.Optional[typing.List[FileInfo]] = pydantic.Field(default=None)
    """
    The list of files downloaded by the browser session
    """

    recordings: typing.Optional[typing.List[FileInfo]] = pydantic.Field(default=None)
    """
    The list of video recordings from the browser session
    """

    started_at: typing.Optional[dt.datetime] = pydantic.Field(default=None)
    """
    Timestamp when the session was started
    """

    completed_at: typing.Optional[dt.datetime] = pydantic.Field(default=None)
    """
    Timestamp when the session was completed
    """

    created_at: dt.datetime = pydantic.Field()
    """
    Timestamp when the session was created (the timestamp for the initial request)
    """

    modified_at: dt.datetime = pydantic.Field()
    """
    Timestamp when the session was last modified
    """

    deleted_at: typing.Optional[dt.datetime] = pydantic.Field(default=None)
    """
    Timestamp when the session was deleted, if applicable
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



"""add_engine_to_workflow_run_block

Revision ID: 2be3e0ba85ff
Revises: 2c6b27e8e961
Create Date: 2025-06-17 07:23:13.753617+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "2be3e0ba85ff"
down_revision: Union[str, None] = "2c6b27e8e961"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("workflow_run_blocks", sa.Column("engine", sa.String(), nullable=True))
    op.create_index(op.f("ix_workflow_run_blocks_task_id"), "workflow_run_blocks", ["task_id"], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_workflow_run_blocks_task_id"), table_name="workflow_run_blocks")
    op.drop_column("workflow_run_blocks", "engine")
    # ### end Alembic commands ###



from datetime import datetime

from pydantic import BaseModel, ConfigDict, Field, model_validator


class BrowserProfile(BaseModel):
    model_config = ConfigDict(from_attributes=True)

    browser_profile_id: str
    organization_id: str
    name: str
    description: str | None = None
    created_at: datetime
    modified_at: datetime
    deleted_at: datetime | None = None


class CreateBrowserProfileRequest(BaseModel):
    name: str = Field(..., description="Name for the browser profile")
    description: str | None = Field(None, description="Optional profile description")
    browser_session_id: str | None = Field(
        default=None, description="Persistent browser session to convert into a profile"
    )
    workflow_run_id: str | None = Field(
        default=None, description="Workflow run whose persisted session should be captured"
    )

    @model_validator(mode="after")
    def _validate_source(self) -> "CreateBrowserProfileRequest":
        if bool(self.browser_session_id) == bool(self.workflow_run_id):
            raise ValueError("Provide either browser_session_id or workflow_run_id")
        return self



"""add ai_fallback field to workflows table

Revision ID: d3ec63728c2a
Revises: 1bba8a38ddc7
Create Date: 2025-08-28 21:12:54.750395+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "d3ec63728c2a"
down_revision: Union[str, None] = "1bba8a38ddc7"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("workflows", sa.Column("ai_fallback", sa.Boolean(), nullable=False, server_default=sa.false()))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workflows", "ai_fallback")
    # ### end Alembic commands ###



"""
A channel for connecting to a persistent browser instance.

What this channel looks like:

    [API Server] <--> [Browser (CDP)]

Channel data:

    CDP protocol data, with Playwright thrown in.
"""

from __future__ import annotations

import asyncio
import functools
import pathlib
import typing as t

import structlog
from playwright.async_api import Browser, BrowserContext, Page, Playwright, async_playwright

from skyvern.config import settings

if t.TYPE_CHECKING:
    from skyvern.forge.sdk.routes.streaming.channels.vnc import VncChannel

LOG = structlog.get_logger()


class CdpChannel:
    """
    CdpChannel. Relies on a VncChannel - without one, a CdpChannel has no
    r'aison d'etre.
    """

    def __new__(cls, *_: t.Iterable[t.Any], **__: t.Mapping[str, t.Any]) -> t.Self:  # noqa: N805
        if cls is CdpChannel:
            raise TypeError("CdpChannel class cannot be instantiated directly.")

        return super().__new__(cls)

    def __init__(self, *, vnc_channel: VncChannel) -> None:
        self.vnc_channel = vnc_channel
        # --
        self.browser: Browser | None = None
        self.browser_context: BrowserContext | None = None
        self.page: Page | None = None
        self.pw: Playwright | None = None
        self.url: str | None = None

    @property
    def class_name(self) -> str:
        return self.__class__.__name__

    @property
    def identity(self) -> t.Dict[str, t.Any]:
        base = self.vnc_channel.identity

        return base | {"url": self.url}

    async def connect(self, cdp_url: str | None = None) -> t.Self:
        """
        Idempotent.
        """

        if self.browser and self.browser.is_connected():
            return self

        await self.close()

        if cdp_url:
            url = cdp_url
        elif self.vnc_channel.browser_session and self.vnc_channel.browser_session.browser_address:
            url = self.vnc_channel.browser_session.browser_address
        else:
            url = settings.BROWSER_REMOTE_DEBUGGING_URL

        self.url = url

        LOG.info(f"{self.class_name} connecting to CDP", **self.identity)

        pw = self.pw or await async_playwright().start()

        self.pw = pw

        headers = (
            {
                "x-api-key": self.vnc_channel.x_api_key,
            }
            if self.vnc_channel.x_api_key
            else None
        )

        def on_close() -> None:
            LOG.warning(
                f"{self.class_name} closing because the persistent browser disconnected itself.", **self.identity
            )
            close_task = asyncio.create_task(self.close())
            close_task.add_done_callback(lambda _: asyncio.create_task(self.connect()))  # TODO: avoid blind reconnect

        self.browser = await pw.chromium.connect_over_cdp(url, headers=headers)
        self.browser.on("disconnected", on_close)

        await self.apply_download_behavior(self.browser)

        contexts = self.browser.contexts
        if contexts:
            LOG.info(f"{self.class_name} using existing browser context", **self.identity)
            self.browser_context = contexts[0]
        else:
            LOG.warning(f"{self.class_name} No existing browser context found, creating new one", **self.identity)
            self.browser_context = await self.browser.new_context()

        pages = self.browser_context.pages
        if pages:
            self.page = pages[0]
            LOG.info(f"{self.class_name} connected to page", **self.identity)
        else:
            LOG.warning(f"{self.class_name} No pages found in browser context", **self.identity)

        LOG.info(f"{self.class_name} connected successfully", **self.identity)

        return self

    async def apply_download_behavior(self, browser: Browser) -> t.Self:
        org_id = self.vnc_channel.organization_id

        browser_session_id = (
            self.vnc_channel.browser_session.persistent_browser_session_id if self.vnc_channel.browser_session else None
        )

        download_path = f"/app/downloads/{org_id}/{browser_session_id}" if browser_session_id else "/app/downloads/"

        cdp_session = await browser.new_browser_cdp_session()

        await cdp_session.send(
            "Browser.setDownloadBehavior",
            {
                "behavior": "allow",
                "downloadPath": download_path,
                "eventsEnabled": True,
            },
        )

        await cdp_session.detach()

        return self

    async def close(self) -> None:
        LOG.info(f"{self.class_name} closing connection", **self.identity)

        if self.browser:
            try:
                await self.browser.close()
            except Exception:
                pass
            self.browser = None

        if self.pw:
            await self.pw.stop()
            self.pw = None

        self.browser_context = None
        self.page = None

        LOG.info(f"{self.class_name} closed", **self.identity)

    async def evaluate_js(
        self,
        expression: str,
        arg: str | int | float | bool | list | dict | None = None,
    ) -> str | int | float | bool | list | dict | None:
        await self.connect()

        if not self.page:
            raise RuntimeError(f"{self.class_name} evaluate_js: not connected to a page. Call connect() first.")

        LOG.info(f"{self.class_name} evaluating js", expression=expression[:100], **self.identity)

        try:
            result = await self.page.evaluate(expression, arg)
            LOG.info(f"{self.class_name} evaluated js successfully", **self.identity)
            return result
        except Exception:
            LOG.exception(f"{self.class_name} failed to evaluate js", expression=expression, **self.identity)
            raise

    @functools.lru_cache(maxsize=None)
    def js(self, file_name: str) -> str:
        base_path = pathlib.Path(__file__).parent / "js"
        file_name = file_name.lstrip("/")

        if not file_name.endswith(".js"):
            file_name += ".js"

        relative_path = pathlib.Path(file_name)
        full_path = base_path / relative_path

        with open(full_path, encoding="utf-8") as f:
            return f.read()



import json
from typing import Any

import curlparser
import structlog

LOG = structlog.get_logger()


def parse_curl_command(curl_command: str) -> dict[str, Any]:
    """
    Parse a curl command into HTTP request parameters using curlparser library.

    Args:
        curl_command: The curl command string to parse

    Returns:
        Dict containing:
        - method: HTTP method (GET, POST, etc.)
        - url: The URL to request
        - headers: Dict of headers
        - body: Request body as dict (parsed from JSON) or None

    Raises:
        ValueError: If the curl command cannot be parsed
    """
    try:
        # Parse the curl command
        parsed = curlparser.parse(curl_command)

        # Extract the components
        method = parsed.method.upper() if parsed.method else "GET"
        if not parsed.method:
            LOG.info(
                "No HTTP method found in curl command, defaulting to GET",
                curl_command=curl_command,
            )

        result = {
            "method": method,
            "url": parsed.url,
            "headers": {},
            "body": None,
        }

        # Process headers - curlparser returns headers as an OrderedDict
        if parsed.header:
            result["headers"] = {k: v.strip() for k, v in parsed.header.items()}

        # Process body/data
        if parsed.data:
            # Try to parse as JSON
            try:
                if isinstance(parsed.data, list):
                    # Join multiple data parts
                    data_str = "".join(parsed.data)
                else:
                    data_str = parsed.data

                result["body"] = json.loads(data_str)
            except (json.JSONDecodeError, TypeError) as e:
                # If not valid JSON, convert to dict with single "data" key
                LOG.warning(
                    "Curl data is not valid JSON, wrapping in data key",
                    data=parsed.data,
                    data_str=data_str if "data_str" in locals() else None,
                    error_type=type(e).__name__,
                    error_message=str(e),
                    curl_command=curl_command,
                )
                result["body"] = {"data": data_str if "data_str" in locals() else parsed.data}

        # Process JSON data if provided
        if hasattr(parsed, "json") and parsed.json:
            try:
                result["body"] = json.loads(parsed.json)
            except (json.JSONDecodeError, TypeError) as e:
                LOG.warning(
                    "Curl json is not valid JSON",
                    json=parsed.json,
                    error_type=type(e).__name__,
                    error_message=str(e),
                    curl_command=curl_command,
                )
                result["body"] = {"data": parsed.json}

        # Validate URL
        if not result["url"]:
            raise ValueError("No URL found in curl command")

        return result

    except Exception as e:
        LOG.exception(
            "Failed to parse curl command",
            curl_command=curl_command,  # Log entire command for debugging
        )
        raise ValueError(f"Failed to parse curl command: {str(e)}")


def curl_to_http_request_block_params(curl_command: str) -> dict[str, Any]:
    """
    Convert a curl command to parameters suitable for HttpRequestBlock.

    This is a convenience function that can be used in API endpoints
    to convert curl commands before creating workflow blocks.

    Args:
        curl_command: The curl command string

    Returns:
        Dict with keys matching HttpRequestBlock parameters
    """
    parsed = parse_curl_command(curl_command)

    return {
        "method": parsed["method"],
        "url": parsed["url"],
        "headers": parsed["headers"] if parsed["headers"] else None,
        "body": parsed["body"],
        "timeout": 30,  # Default timeout
        "follow_redirects": True,  # Default follow redirects
    }



from __future__ import annotations

import asyncio
import os
import time
from collections import OrderedDict
from contextvars import ContextVar, Token
from dataclasses import dataclass
from threading import RLock

import structlog
from fastapi import HTTPException
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.types import ASGIApp, Receive, Scope, Send

from skyvern.config import settings
from skyvern.forge.sdk.db.agent_db import AgentDB
from skyvern.forge.sdk.db.enums import OrganizationAuthTokenType
from skyvern.forge.sdk.services.org_auth_service import resolve_org_from_api_key

from .api_key_hash import hash_api_key_for_cache
from .client import reset_api_key_override, set_api_key_override

LOG = structlog.get_logger(__name__)
API_KEY_HEADER = "x-api-key"
TARGET_ORG_ID_HEADER = "x-target-org-id"
HEALTH_PATHS = {"/health", "/healthz"}
_MCP_ALLOWED_TOKEN_TYPES = (
    OrganizationAuthTokenType.api,
    OrganizationAuthTokenType.mcp_admin_impersonation,
)
_auth_db: AgentDB | None = None
_auth_db_lock = RLock()
_api_key_cache_lock = RLock()
_api_key_validation_cache: OrderedDict[str, tuple[MCPAPIKeyValidation | None, float]] = OrderedDict()
_NEGATIVE_CACHE_TTL_SECONDS = 5.0
_VALIDATION_RETRY_EXHAUSTED_MESSAGE = "API key validation temporarily unavailable"
_MAX_VALIDATION_RETRIES = 2
_RETRY_DELAY_SECONDS = 0.25

# ---------------------------------------------------------------------------
# Impersonation session state
# ---------------------------------------------------------------------------

_admin_api_key_hash: ContextVar[str | None] = ContextVar("admin_api_key_hash", default=None)


@dataclass(frozen=True)
class ImpersonationSession:
    admin_api_key_hash: str
    admin_org_id: str
    target_org_id: str
    # Stored in plaintext in process memory for up to TTL. Acceptable V1 trade-off
    # (in-process only, same as API key cache).
    target_api_key: str
    expires_at: float  # time.monotonic() deadline
    ttl_minutes: int


_impersonation_sessions: dict[str, ImpersonationSession] = {}
_impersonation_lock = RLock()

MAX_TTL_MINUTES = 120
DEFAULT_TTL_MINUTES = 30
_MAX_IMPERSONATION_SESSIONS = 100


def get_active_impersonation(admin_key_hash: str) -> ImpersonationSession | None:
    with _impersonation_lock:
        session = _impersonation_sessions.get(admin_key_hash)
        if session is None:
            return None
        if session.expires_at <= time.monotonic():
            _impersonation_sessions.pop(admin_key_hash, None)
            return None
        return session


def set_impersonation_session(session: ImpersonationSession) -> None:
    with _impersonation_lock:
        _impersonation_sessions[session.admin_api_key_hash] = session
        # Sweep expired sessions, then evict oldest if over capacity
        now = time.monotonic()
        expired_keys = [k for k, s in _impersonation_sessions.items() if s.expires_at <= now]
        for k in expired_keys:
            _impersonation_sessions.pop(k, None)
        if len(_impersonation_sessions) > _MAX_IMPERSONATION_SESSIONS:
            oldest_key = min(_impersonation_sessions, key=lambda k: _impersonation_sessions[k].expires_at)
            _impersonation_sessions.pop(oldest_key, None)


def clear_impersonation_session(admin_key_hash: str) -> ImpersonationSession | None:
    with _impersonation_lock:
        return _impersonation_sessions.pop(admin_key_hash, None)


def clear_all_impersonation_sessions() -> None:
    with _impersonation_lock:
        _impersonation_sessions.clear()


def get_admin_api_key_hash() -> str | None:
    return _admin_api_key_hash.get()


# ---------------------------------------------------------------------------


@dataclass(frozen=True)
class MCPAPIKeyValidation:
    organization_id: str
    token_type: OrganizationAuthTokenType


def _resolve_api_key_cache_ttl_seconds() -> float:
    raw = os.environ.get("SKYVERN_MCP_API_KEY_CACHE_TTL_SECONDS", "30")
    try:
        return max(1.0, float(raw))
    except ValueError:
        return 30.0


def _resolve_api_key_cache_max_size() -> int:
    raw = os.environ.get("SKYVERN_MCP_API_KEY_CACHE_MAX_SIZE", "1024")
    try:
        return max(1, int(raw))
    except ValueError:
        return 1024


_API_KEY_CACHE_TTL_SECONDS = _resolve_api_key_cache_ttl_seconds()
_API_KEY_CACHE_MAX_SIZE = _resolve_api_key_cache_max_size()


def _get_auth_db() -> AgentDB:
    global _auth_db
    # Guard singleton init in case HTTP transport is served with threaded workers.
    with _auth_db_lock:
        if _auth_db is None:
            # Prefer CloudAgentDB when available (cloud deploys) because the base
            # AgentDB.validate_org_auth_token does not handle encrypted tokens,
            # causing all API key validation to fail when ENABLE_ENCRYPTION is on.
            try:
                from cloud.db.cloud_agent_db import CloudAgentDB  # noqa: PLC0415

                _auth_db = CloudAgentDB(settings.DATABASE_STRING, debug_enabled=settings.DEBUG_MODE)
                LOG.info("MCP auth DB initialized", db_class="CloudAgentDB")
            except ImportError:
                _auth_db = AgentDB(settings.DATABASE_STRING, debug_enabled=settings.DEBUG_MODE)
                LOG.info("MCP auth DB initialized", db_class="AgentDB")
    return _auth_db


async def close_auth_db() -> None:
    """Dispose the auth DB engine used by HTTP middleware, if initialized."""
    global _auth_db
    with _auth_db_lock:
        db = _auth_db
        _auth_db = None
    with _api_key_cache_lock:
        _api_key_validation_cache.clear()
    clear_all_impersonation_sessions()
    if db is None:
        return

    try:
        await db.engine.dispose()
    except Exception:
        LOG.warning("Failed to dispose MCP auth DB engine", exc_info=True)


def cache_key(api_key: str) -> str:
    return hash_api_key_for_cache(api_key)


def _admin_organization_ids() -> set[str]:
    try:
        from cloud.config import settings as cloud_settings  # noqa: PLC0415

        admin_ids = getattr(cloud_settings, "ADMIN_ORGANIZATION_IDS", [])
    except ImportError:
        admin_ids = getattr(settings, "ADMIN_ORGANIZATION_IDS", [])
    return {org_id for org_id in admin_ids if org_id}


def _is_admin_impersonation_enabled() -> bool:
    try:
        from cloud.config import settings as cloud_settings  # noqa: PLC0415

        return bool(getattr(cloud_settings, "MCP_ADMIN_IMPERSONATION_ENABLED", False))
    except ImportError:
        return bool(getattr(settings, "MCP_ADMIN_IMPERSONATION_ENABLED", False))


async def validate_mcp_api_key(api_key: str) -> MCPAPIKeyValidation:
    """Validate API key and return caller organization + token type."""
    key = cache_key(api_key)

    # Check cache first.
    with _api_key_cache_lock:
        cached = _api_key_validation_cache.get(key)
        if cached is not None:
            cached_validation, expires_at = cached
            if expires_at > time.monotonic():
                _api_key_validation_cache.move_to_end(key)
                if cached_validation is None:
                    raise HTTPException(status_code=401, detail="Invalid API key")
                return cached_validation
            _api_key_validation_cache.pop(key, None)

    # Cache miss — do the DB lookup with simple retry on transient errors.
    last_exc: Exception | None = None
    for attempt in range(_MAX_VALIDATION_RETRIES + 1):
        if attempt > 0:
            await asyncio.sleep(_RETRY_DELAY_SECONDS)
        try:
            validation = await resolve_org_from_api_key(
                api_key,
                _get_auth_db(),
                token_types=_MCP_ALLOWED_TOKEN_TYPES,
            )
            caller_validation = MCPAPIKeyValidation(
                organization_id=validation.organization.organization_id,
                token_type=validation.token.token_type,
            )
            with _api_key_cache_lock:
                _api_key_validation_cache[key] = (
                    caller_validation,
                    time.monotonic() + _API_KEY_CACHE_TTL_SECONDS,
                )
                _api_key_validation_cache.move_to_end(key)
                while len(_api_key_validation_cache) > _API_KEY_CACHE_MAX_SIZE:
                    _api_key_validation_cache.popitem(last=False)
            return caller_validation
        except HTTPException as e:
            if e.status_code in {401, 403}:
                with _api_key_cache_lock:
                    _api_key_validation_cache[key] = (None, time.monotonic() + _NEGATIVE_CACHE_TTL_SECONDS)
                    _api_key_validation_cache.move_to_end(key)
                    while len(_api_key_validation_cache) > _API_KEY_CACHE_MAX_SIZE:
                        _api_key_validation_cache.popitem(last=False)
                raise
            last_exc = e
        except Exception as e:
            last_exc = e

    LOG.warning("API key validation retries exhausted", attempts=_MAX_VALIDATION_RETRIES + 1, exc_info=last_exc)
    raise HTTPException(status_code=503, detail=_VALIDATION_RETRY_EXHAUSTED_MESSAGE)


def _unauthorized_response(message: str) -> JSONResponse:
    return JSONResponse({"error": {"code": "UNAUTHORIZED", "message": message}}, status_code=401)


def _internal_error_response() -> JSONResponse:
    return JSONResponse(
        {"error": {"code": "INTERNAL_ERROR", "message": "Internal server error"}},
        status_code=500,
    )


def _service_unavailable_response(message: str) -> JSONResponse:
    return JSONResponse(
        {"error": {"code": "SERVICE_UNAVAILABLE", "message": message}},
        status_code=503,
    )


def _deny_impersonation(
    *,
    reason: str,
    caller_organization_id: str | None,
    target_organization_id: str | None,
    token_type: OrganizationAuthTokenType | None = None,
) -> JSONResponse:
    log_kwargs: dict[str, object] = {
        "reason": reason,
        "caller_organization_id": caller_organization_id,
        "target_organization_id": target_organization_id,
    }
    if token_type is not None:
        log_kwargs["token_type"] = token_type.value
    LOG.warning("MCP admin impersonation denied", **log_kwargs)
    return _unauthorized_response("Impersonation not allowed")


class MCPAPIKeyMiddleware:
    """Require x-api-key for MCP HTTP transport and scope requests to that key."""

    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        request = Request(scope, receive=receive)
        if request.url.path in HEALTH_PATHS:
            response = JSONResponse({"status": "ok"})
            await response(scope, receive, send)
            return

        if request.method == "OPTIONS":
            await self.app(scope, receive, send)
            return

        api_key = request.headers.get(API_KEY_HEADER)
        if not api_key:
            response = _unauthorized_response("Missing x-api-key header")
            await response(scope, receive, send)
            return

        target_org_id_header = request.headers.get(TARGET_ORG_ID_HEADER)
        admin_hash_token: Token[str | None] | None = None

        try:
            validation = await validate_mcp_api_key(api_key)
            caller_organization_id = validation.organization_id
            admin_key_hash = cache_key(api_key)

            scope.setdefault("state", {})
            if target_org_id_header is not None:
                # Explicit header-based impersonation (takes priority over session)
                if not _is_admin_impersonation_enabled():
                    response = _deny_impersonation(
                        reason="feature_disabled",
                        caller_organization_id=caller_organization_id,
                        target_organization_id=target_org_id_header.strip() or target_org_id_header,
                    )
                    await response(scope, receive, send)
                    return

                target_organization_id = target_org_id_header.strip()
                if not target_organization_id:
                    response = _deny_impersonation(
                        reason="missing_target_organization_id",
                        caller_organization_id=caller_organization_id,
                        target_organization_id=target_org_id_header,
                        token_type=validation.token_type,
                    )
                    await response(scope, receive, send)
                    return

                # Delegate validation to the single source of truth in cloud/.
                # The import can't fail here — _is_admin_impersonation_enabled()
                # already returned True, which requires cloud.config to be importable.
                try:
                    from cloud.mcp_admin_tools import validate_impersonation_target  # noqa: PLC0415
                except ImportError:
                    response = _deny_impersonation(
                        reason="impersonation_not_available",
                        caller_organization_id=caller_organization_id,
                        target_organization_id=target_organization_id,
                    )
                    await response(scope, receive, send)
                    return

                result = await validate_impersonation_target(
                    caller_organization_id=caller_organization_id,
                    target_organization_id=target_organization_id,
                    token_type=validation.token_type,
                )
                if isinstance(result, str):
                    response = _deny_impersonation(
                        reason=result,
                        caller_organization_id=caller_organization_id,
                        target_organization_id=target_organization_id,
                        token_type=validation.token_type,
                    )
                    await response(scope, receive, send)
                    return

                resolved_org_id, target_api_key = result
                api_key = target_api_key

                scope["state"]["organization_id"] = resolved_org_id
                scope["state"]["admin_organization_id"] = caller_organization_id
                scope["state"]["impersonation_target_organization_id"] = resolved_org_id
                admin_hash_token = _admin_api_key_hash.set(admin_key_hash)
                LOG.info(
                    "MCP admin impersonation allowed",
                    caller_organization_id=caller_organization_id,
                    target_organization_id=resolved_org_id,
                    token_type=validation.token_type.value,
                )
            else:
                # No explicit header — check for session-based impersonation
                session = get_active_impersonation(admin_key_hash)
                if session is not None:
                    # Session data (target API key) is cached for TTL. If the target org's
                    # key is revoked mid-session, impersonation continues until expiry —
                    # acceptable trade-off vs re-validating every request.
                    api_key = session.target_api_key
                    scope["state"]["organization_id"] = session.target_org_id
                    scope["state"]["admin_organization_id"] = session.admin_org_id
                    scope["state"]["impersonation_target_organization_id"] = session.target_org_id
                    admin_hash_token = _admin_api_key_hash.set(admin_key_hash)
                    LOG.info(
                        "MCP session impersonation applied",
                        caller_organization_id=session.admin_org_id,
                        target_organization_id=session.target_org_id,
                        ttl_minutes=session.ttl_minutes,
                    )
                else:
                    scope["state"]["organization_id"] = caller_organization_id
        except HTTPException as e:
            if e.status_code in {401, 403}:
                response = _unauthorized_response("Invalid API key")
                await response(scope, receive, send)
                return
            if e.status_code == 503:
                response = _service_unavailable_response(e.detail or _VALIDATION_RETRY_EXHAUSTED_MESSAGE)
                await response(scope, receive, send)
                return
            LOG.warning("Unexpected HTTPException during MCP API key validation", status_code=e.status_code)
            response = _internal_error_response()
            await response(scope, receive, send)
            return
        except Exception:
            LOG.exception("Unexpected MCP API key validation failure")
            response = _internal_error_response()
            await response(scope, receive, send)
            return

        token = set_api_key_override(api_key)
        try:
            await self.app(scope, receive, send)
        finally:
            reset_api_key_override(token)
            if admin_hash_token is not None:
                _admin_api_key_hash.reset(admin_hash_token)



# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel, update_forward_refs


class FileNode(UniversalBaseModel):
    """
    Model representing a file or directory in the file tree.
    """

    type: str = pydantic.Field()
    """
    Type of node: 'file' or 'directory'
    """

    size: typing.Optional[int] = pydantic.Field(default=None)
    """
    File size in bytes
    """

    mime_type: typing.Optional[str] = pydantic.Field(default=None)
    """
    MIME type of the file
    """

    content_hash: typing.Optional[str] = pydantic.Field(default=None)
    """
    SHA256 hash of file content
    """

    created_at: dt.datetime = pydantic.Field()
    """
    Timestamp when the file was created
    """

    children: typing.Optional[typing.Dict[str, typing.Optional["FileNode"]]] = pydantic.Field(default=None)
    """
    Child nodes for directories
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


update_forward_refs(FileNode)



import re

from jinja2 import Environment, StrictUndefined, UndefinedError, meta


class Constants:
    MissingVariablePattern = var_pattern = r"\{\{\s*([a-zA-Z_][a-zA-Z0-9_.\[\]'\"]*)\s*\}\}"


def replace_jinja_reference(text: str, old_key: str, new_key: str) -> str:
    """Replaces jinja-style references in a string.

    Handles patterns like {{oldKey}}, {{oldKey.field}}, {{oldKey | filter}}, {{oldKey[0]}}

    Args:
        text: The text to search in
        old_key: The key to replace (without braces)
        new_key: The new key to use (without braces)

    Returns:
        The text with references replaced
    """
    # Match {{oldKey}} or {{oldKey.something}} or {{oldKey | filter}} or {{oldKey[0]}} etc.
    # Use negative lookahead to ensure key is not followed by identifier characters,
    # which prevents matching {{keyOther}} when searching for {{key}}
    # Capture whitespace after {{ to preserve formatting (e.g., "{{ key }}" stays "{{ newKey }}")
    escaped_old_key = re.escape(old_key)
    pattern = rf"\{{\{{(\s*){escaped_old_key}(?![a-zA-Z0-9_])"
    return re.sub(pattern, rf"{{{{\1{new_key}", text)


def get_missing_variables(template_source: str, template_data: dict) -> set[str]:
    # quick check - catch top-level undefineds
    env = Environment(undefined=StrictUndefined)
    ast = env.parse(template_source)
    undeclared_vars = meta.find_undeclared_variables(ast)
    missing_vars = undeclared_vars - set(template_data.keys())

    # nested undefined won't be caught; let's check for those
    if not missing_vars:
        # try rendering to catch nested undefineds (dotted attributes, list/dict access)
        try:
            template = env.from_string(template_source)
            template.render(template_data)
        except UndefinedError:
            # matches: {{ var }}, {{ var.attr }}, {{ var[0] }}, {{ var['key'] }}, {{ var.attr[0] }}
            matches = re.findall(Constants.MissingVariablePattern, template_source)

            for match in matches:
                root = match.split("[")[0].split(".")[0]

                # just check if the 'root' of the variable exists in the provided data
                # if it does, add the whole match as missing
                if root in template_data:
                    missing_vars.add(match)

            if not missing_vars:
                raise  # re-raise if we couldn't determine missing vars

    return missing_vars



from __future__ import annotations

import asyncio
import os
from pathlib import Path
from typing import Any, Callable

import pyotp
import structlog
from cachetools import TTLCache
from playwright.async_api import Page

from skyvern.config import settings
from skyvern.constants import BROWSER_DOWNLOAD_TIMEOUT
from skyvern.core.script_generations.real_skyvern_page_ai import RealSkyvernPageAi, render_template
from skyvern.core.script_generations.skyvern_page import ActionCall, ActionMetadata, RunContext, SkyvernPage
from skyvern.core.script_generations.skyvern_page_ai import SkyvernPageAi
from skyvern.errors.errors import UserDefinedError
from skyvern.exceptions import ScriptTerminationException, WorkflowRunNotFound
from skyvern.forge import app
from skyvern.forge.prompts import prompt_engine
from skyvern.forge.sdk.api.files import (
    check_downloading_files_and_wait_for_download_to_complete,
    get_path_for_workflow_download_directory,
    list_files_in_directory,
)
from skyvern.forge.sdk.artifact.models import ArtifactType
from skyvern.forge.sdk.core import skyvern_context
from skyvern.schemas.steps import AgentStepOutput
from skyvern.services.otp_service import poll_otp_value, try_generate_totp_from_credential
from skyvern.utils.url_validators import prepend_scheme_and_validate_url
from skyvern.webeye.actions.action_types import ActionType
from skyvern.webeye.actions.actions import (
    Action,
    ActionStatus,
    CompleteAction,
    DecisiveAction,
    ExtractAction,
    SelectOption,
    SolveCaptchaAction,
)
from skyvern.webeye.actions.handler import (
    ActionHandler,
    generate_totp_value,
    get_actual_value_of_parameter_if_secret,
    handle_complete_action,
)
from skyvern.webeye.actions.responses import ActionFailure, ActionResult, ActionSuccess
from skyvern.webeye.browser_state import BrowserState
from skyvern.webeye.scraper.scraped_page import ScrapedPage
from skyvern.webeye.utils.page import SkyvernFrame

LOG = structlog.get_logger()

action_wrap = SkyvernPage.action_wrap


class ScriptSkyvernPage(SkyvernPage):
    """
    A minimal adapter around the chosen driver that:
    1. Executes real browser commands
    2. Records ActionCallobjects into RunContext.trace
    3. Adds retry / fallback hooks
    """

    def __init__(
        self,
        scraped_page: ScrapedPage,
        page: Page,
        ai: SkyvernPageAi,
        *,
        recorder: Callable[[ActionCall], None] | None = None,
    ) -> None:
        super().__init__(page=page, ai=ai)
        self.scraped_page = scraped_page
        self._record = recorder or (lambda ac: None)

    @classmethod
    async def _get_or_create_browser_state(cls, browser_session_id: str | None = None) -> BrowserState:
        context = skyvern_context.current()
        if context and context.workflow_run_id and context.organization_id:
            workflow_run = await app.DATABASE.get_workflow_run(
                workflow_run_id=context.workflow_run_id, organization_id=context.organization_id
            )
            if workflow_run:
                browser_state = await app.BROWSER_MANAGER.get_or_create_for_workflow_run(
                    workflow_run=workflow_run,
                    browser_session_id=browser_session_id,
                    browser_profile_id=workflow_run.browser_profile_id,
                )
            else:
                raise WorkflowRunNotFound(workflow_run_id=context.workflow_run_id)
        else:
            browser_state = await app.BROWSER_MANAGER.get_or_create_for_script(browser_session_id=browser_session_id)
        return browser_state

    @classmethod
    async def _get_browser_state(cls) -> BrowserState | None:
        context = skyvern_context.current()
        if context and context.workflow_run_id and context.organization_id:
            workflow_run = await app.DATABASE.get_workflow_run(
                workflow_run_id=context.workflow_run_id, organization_id=context.organization_id
            )
            if workflow_run:
                browser_state = app.BROWSER_MANAGER.get_for_workflow_run(workflow_run_id=context.workflow_run_id)
            else:
                raise WorkflowRunNotFound(workflow_run_id=context.workflow_run_id)
        else:
            browser_state = app.BROWSER_MANAGER.get_for_script()
        return browser_state

    @classmethod
    async def create(
        cls,
        browser_session_id: str | None = None,
    ) -> ScriptSkyvernPage:
        scraped_page = await cls.create_scraped_page(browser_session_id=browser_session_id)
        page = await scraped_page._browser_state.must_get_working_page()
        ai = RealSkyvernPageAi(scraped_page, page)
        return cls(scraped_page=scraped_page, page=page, ai=ai)

    @classmethod
    async def create_scraped_page(
        cls,
        browser_session_id: str | None = None,
    ) -> ScrapedPage:
        # initialize browser state
        # TODO: add workflow_run_id or eventually script_id/script_run_id
        browser_state = await cls._get_or_create_browser_state(browser_session_id=browser_session_id)
        return await browser_state.scrape_website(
            url="",
            cleanup_element_tree=app.AGENT_FUNCTION.cleanup_element_tree_factory(),
            scrape_exclude=app.scrape_exclude,
            max_screenshot_number=settings.MAX_NUM_SCREENSHOTS,
            draw_boxes=True,
            scroll=True,
            support_empty_page=True,
        )

    async def _ensure_download_to_complete(
        self,
        download_dir: Path,
        browser_session_id: str | None = None,
    ) -> None:
        context = skyvern_context.current()
        if not context or not context.organization_id:
            return
        if not download_dir.exists():
            return
        organization_id = context.organization_id
        download_timeout = BROWSER_DOWNLOAD_TIMEOUT
        if context.task_id:
            task = await app.DATABASE.get_task(context.task_id, organization_id=organization_id)
            if task and task.download_timeout:
                download_timeout = task.download_timeout
        await check_downloading_files_and_wait_for_download_to_complete(
            download_dir=download_dir,
            organization_id=organization_id,
            browser_session_id=browser_session_id,
            timeout=download_timeout,
        )

    async def _decorate_call(
        self,
        fn: Callable,
        action: ActionType,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """
        Decorator to record the action call.

        Auto-creates action records in DB before action execution
        and screenshot artifacts after action execution.
        """

        # Emoji mapping for different action types
        ACTION_EMOJIS = {
            ActionType.CLICK: "👆",
            ActionType.INPUT_TEXT: "⌨️",
            ActionType.UPLOAD_FILE: "📤",
            ActionType.DOWNLOAD_FILE: "📥",
            ActionType.HOVER: "🖱️",
            ActionType.SELECT_OPTION: "🎯",
            ActionType.WAIT: "⏳",
            ActionType.SOLVE_CAPTCHA: "🔓",
            ActionType.VERIFICATION_CODE: "🔐",
            ActionType.SCROLL: "📜",
            ActionType.COMPLETE: "✅",
            ActionType.TERMINATE: "🛑",
        }

        prompt = kwargs.get("prompt", "")

        # Backward compatibility: use intention if provided and prompt is empty
        intention = kwargs.get("intention", None)
        if intention and not prompt:
            prompt = intention

        data = kwargs.get("data", None)
        meta = ActionMetadata(prompt, data)
        call = ActionCall(action, args, kwargs, meta)

        action_status = ActionStatus.completed

        # Print action in script mode
        context = skyvern_context.current()
        if context and context.script_mode:
            emoji = ACTION_EMOJIS.get(action, "🔧")
            action_name = action.value if hasattr(action, "value") else str(action)
            print(f"{emoji} {action_name.replace('_', ' ').title()}", end="")
            if prompt:
                print(f": {prompt}")
            else:
                print()

        # Download detection for click actions
        download_triggered: bool | None = None
        downloaded_files: list[str] | None = None
        files_before: list[str] = []
        download_dir: Path | None = None

        # Capture files before click action for download detection
        if action == ActionType.CLICK and context and context.workflow_run_id:
            try:
                download_dir = get_path_for_workflow_download_directory(context.workflow_run_id)
                if download_dir.exists():
                    files_before = list_files_in_directory(download_dir)
                if context.browser_session_id and context.organization_id:
                    browser_session_downloaded_files = await app.STORAGE.list_downloaded_files_in_browser_session(
                        organization_id=context.organization_id,
                        browser_session_id=context.browser_session_id,
                    )
                    files_before = files_before + browser_session_downloaded_files

            except Exception:
                pass  # Don't block action execution if file listing fails

        try:
            # Wait for page to be ready before executing action
            # This helps prevent issues where cached actions execute before the page is fully loaded
            await self._wait_for_page_ready_before_action()
            await self._ensure_element_ids_on_page()

            call.result = await fn(self, *args, **kwargs)

            # Note: Action status would be updated to completed here if update method existed

            # Print success in script mode
            if context and context.script_mode:
                print("  ✓ Completed")

            return call.result
        except Exception as e:
            call.error = e
            action_status = ActionStatus.failed
            # Note: Action status would be updated to failed here if update method existed

            # Print failure in script mode
            if context and context.script_mode:
                print(f"  ✗ Failed: {str(e)}")

            # LLM fallback hook could go here ...
            raise
        finally:
            # Add a small buffer between cached actions to give slow pages time to settle
            if settings.CACHED_ACTION_DELAY_SECONDS > 0:
                await asyncio.sleep(settings.CACHED_ACTION_DELAY_SECONDS)

            # Check for downloaded files after click action
            if action == ActionType.CLICK and context and context.workflow_run_id and download_dir:
                try:
                    if download_dir.exists():
                        await self._ensure_download_to_complete(
                            download_dir=download_dir,
                            browser_session_id=context.browser_session_id,
                        )
                        files_after = list_files_in_directory(download_dir)
                        if context.browser_session_id and context.organization_id:
                            browser_session_downloaded_files = (
                                await app.STORAGE.list_downloaded_files_in_browser_session(
                                    organization_id=context.organization_id,
                                    browser_session_id=context.browser_session_id,
                                )
                            )
                            files_after = files_after + browser_session_downloaded_files

                        new_file_paths = set(files_after) - set(files_before)
                        if new_file_paths:
                            download_triggered = True
                            downloaded_files = [os.path.basename(fp) for fp in new_file_paths]
                            LOG.info(
                                "Script click action detected download",
                                downloaded_files=downloaded_files,
                                workflow_run_id=context.workflow_run_id,
                            )
                        else:
                            download_triggered = False
                except Exception:
                    pass  # Don't block if download detection fails

            self._record(call)
            # Auto-create action after execution and store result
            await self._create_action_and_result_after_execution(
                action_type=action,
                intention=prompt,
                status=action_status,
                kwargs=kwargs,
                call_result=call.result,
                call_error=call.error,
                download_triggered=download_triggered,
                downloaded_files=downloaded_files,
            )

            # Auto-create screenshot artifact after execution
            await self._create_screenshot_after_execution()

    async def _update_action_reasoning(
        self,
        action_id: str,
        organization_id: str,
        action_type: ActionType,
        intention: str = "",
        text: str | None = None,
        select_option: SelectOption | None = None,
        file_url: str | None = None,
        data_extraction_goal: str | None = None,
        data_extraction_schema: dict[str, Any] | list | str | None = None,
    ) -> str:
        """Generate user-facing reasoning for an action using the secondary LLM."""

        reasoning = f"Auto-generated action for {action_type.value}"
        try:
            context = skyvern_context.current()
            if not context or not context.organization_id:
                return f"Auto-generated action for {action_type.value}"

            # Build the prompt with available context
            reasoning_prompt = prompt_engine.load_prompt(
                template="generate-action-reasoning",
                action_type=action_type.value,
                intention=intention,
                text=text,
                select_option=select_option.value if select_option else None,
                file_url=file_url,
                data_extraction_goal=data_extraction_goal,
                data_extraction_schema=data_extraction_schema,
            )

            # Call secondary LLM to generate reasoning
            json_response = await app.SECONDARY_LLM_API_HANDLER(
                prompt=reasoning_prompt,
                prompt_name="generate-action-reasoning",
                organization_id=context.organization_id,
            )

            reasoning = json_response.get("reasoning", f"Auto-generated action for {action_type.value}")

        except Exception:
            LOG.warning("Failed to generate action reasoning, using fallback", action_type=action_type)
        await app.DATABASE.update_action_reasoning(
            organization_id=organization_id,
            action_id=action_id,
            reasoning=reasoning,
        )
        return reasoning

    async def _create_action_and_result_after_execution(
        self,
        action_type: ActionType,
        intention: str = "",
        status: ActionStatus = ActionStatus.pending,
        kwargs: dict[str, Any] | None = None,
        call_result: Any | None = None,
        call_error: Exception | None = None,
        download_triggered: bool | None = None,
        downloaded_files: list[str] | None = None,
    ) -> tuple[Action | None, list[ActionResult]]:
        """Create an action record and result in the database after execution if task_id and step_id are available.

        Returns a tuple of (Action, list[ActionResult]) similar to how the agent stores actions and results.
        """
        results: list[ActionResult] = []

        try:
            context = skyvern_context.current()
            if not context or not context.task_id or not context.step_id:
                return None, results

            # Create action record. TODO: store more action fields
            kwargs = kwargs or {}
            # we're using "value" instead of "text" for input text actions interface
            xpath = None
            if action_type == ActionType.CLICK:
                if isinstance(call_result, str) and "xpath=" in call_result:
                    xpath_split_list = call_result.split("xpath=")
                    if len(xpath_split_list) > 1:
                        xpath = xpath_split_list[1]
            text = None
            select_option = None
            response: str | None = kwargs.get("response")
            file_url = kwargs.get("file_url")
            if not response:
                if action_type == ActionType.INPUT_TEXT:
                    text = str(call_result)
                    response = text
                elif action_type == ActionType.SELECT_OPTION:
                    option_value = str(call_result) or ""
                    select_option = SelectOption(value=option_value)
                    response = option_value
                elif action_type == ActionType.UPLOAD_FILE:
                    file_url = str(call_result)

            action = Action(
                element_id="",
                action_type=action_type,
                status=status,
                organization_id=context.organization_id,
                workflow_run_id=context.workflow_run_id,
                task_id=context.task_id,
                step_id=context.step_id,
                step_order=0,  # Will be updated by the system if needed
                action_order=context.action_order,  # Will be updated by the system if needed
                intention=intention,
                text=text,
                option=select_option,
                file_url=file_url,
                response=response,
                xpath=xpath,
                download=download_triggered,
                download_triggered=download_triggered,
                downloaded_files=downloaded_files,
                created_by="script",
            )
            data_extraction_goal = None
            data_extraction_schema = None
            if action_type == ActionType.EXTRACT:
                data_extraction_goal = kwargs.get("prompt")
                data_extraction_schema = kwargs.get("schema")
                action = ExtractAction(
                    element_id="",
                    action_type=action_type,
                    status=status,
                    organization_id=context.organization_id,
                    workflow_run_id=context.workflow_run_id,
                    task_id=context.task_id,
                    step_id=context.step_id,
                    step_order=0,
                    action_order=context.action_order,
                    intention=intention,
                    data_extraction_goal=data_extraction_goal,
                    data_extraction_schema=data_extraction_schema,
                    option=select_option,
                    response=response,
                    created_by="script",
                )

            created_action = await app.DATABASE.create_action(action)
            # Generate user-facing reasoning using secondary LLM
            asyncio.create_task(
                self._update_action_reasoning(
                    action_id=str(created_action.action_id),
                    organization_id=str(context.organization_id),
                    action_type=action_type,
                    intention=intention,
                    text=text,
                    select_option=select_option,
                    file_url=file_url,
                    data_extraction_goal=data_extraction_goal,
                    data_extraction_schema=data_extraction_schema,
                )
            )

            context.action_order += 1

            # Create ActionResult based on success/failure
            if call_error:
                result = ActionFailure(exception=call_error, download_triggered=download_triggered)
            else:
                # For extract actions, include the extracted data in the result
                result_data = None
                if action_type == ActionType.EXTRACT and call_result:
                    result_data = call_result
                result = ActionSuccess(
                    data=result_data,
                    download_triggered=download_triggered,
                    downloaded_files=downloaded_files,
                )

            results = [result]

            # Store action and results in RunContext for step output
            run_context = script_run_context_manager.get_run_context()
            if run_context:
                run_context.actions_and_results.append((created_action, results))

            return created_action, results

        except Exception:
            # If action creation fails, don't block the actual action execution
            return None, results

    @classmethod
    async def _create_screenshot_after_execution(cls) -> None:
        """Create a screenshot artifact after action execution if task_id and step_id are available."""
        try:
            context = skyvern_context.ensure_context()
            if not context or not context.task_id or not context.step_id:
                return

            # Get browser state and take screenshot
            browser_state = await cls._get_browser_state()
            if not browser_state:
                return

            screenshot = await browser_state.take_post_action_screenshot(scrolling_number=0)

            if screenshot:
                # Create a minimal Step object for artifact creation
                step = await app.DATABASE.get_step(
                    context.step_id,
                    organization_id=context.organization_id,
                )
                if not step:
                    return

                await app.ARTIFACT_MANAGER.create_artifact(
                    step=step,
                    artifact_type=ArtifactType.SCREENSHOT_ACTION,
                    data=screenshot,
                )

        except Exception:
            # If screenshot creation fails, don't block execution
            pass

    async def _wait_for_page_ready_before_action(self) -> None:
        """
        Wait for the page to be ready before executing a cached action.

        This addresses issues like SKY-6814, SKY-7476, SKY-7344 where cached actions
        execute before the page is fully loaded (e.g., after login transitions).

        The method checks for:
        1. Network idle (with short timeout - some pages never go idle)
        2. Loading indicators (spinners, skeletons, progress bars)
        3. DOM stability (no significant mutations for 300ms)
        """
        try:
            # Note: SkyvernPage uses self.page, not self._page
            if not self.page:
                return

            skyvern_frame = await SkyvernFrame.create_instance(frame=self.page)
            await skyvern_frame.wait_for_page_ready(
                network_idle_timeout_ms=settings.PAGE_READY_NETWORK_IDLE_TIMEOUT_MS,
                loading_indicator_timeout_ms=settings.PAGE_READY_LOADING_INDICATOR_TIMEOUT_MS,
                dom_stable_ms=settings.PAGE_READY_DOM_STABLE_MS,
                dom_stability_timeout_ms=settings.PAGE_READY_DOM_STABILITY_TIMEOUT_MS,
            )
        except Exception:
            # Don't block action execution if page readiness check fails
            LOG.debug("Page readiness check failed, proceeding with action", exc_info=True)

    async def _ensure_element_ids_on_page(self) -> None:
        """
        Ensure unique_id attributes exist on DOM elements for cached selectors.

        After page navigation, the new DOM has no unique_id attributes because
        they are only set during scraping (domUtils.js buildTreeFromBody). Cached
        actions use [unique_id='XXX'] selectors, so we need to build the element
        tree before executing cached actions on a new page.
        """
        try:
            if not self.page:
                return

            # Quick check: do unique_id attributes already exist?
            has_unique_ids = await self.page.evaluate("() => document.querySelector('[unique_id]') !== null")
            if has_unique_ids:
                return

            # Inject domUtils.js and build the element tree to set unique_id attrs.
            # Use a short timeout since this is best-effort; we don't want to hang for 60s.
            skyvern_frame = await SkyvernFrame.create_instance(frame=self.page)
            await skyvern_frame.build_tree_from_body(
                frame_name="main.frame",
                frame_index=0,
                timeout_ms=15000,
            )
            LOG.info("Injected element IDs on page for cached script execution")
        except Exception:
            LOG.debug(
                "Failed to ensure element IDs on page, proceeding with action",
                exc_info=True,
            )

    async def get_actual_value(
        self,
        value: str,
        totp_identifier: str | None = None,
        totp_url: str | None = None,
    ) -> str:
        """Input text into an element identified by ``selector``."""
        context = skyvern_context.ensure_context()
        if context and context.workflow_run_id:
            task_id = context.task_id
            workflow_run_id = context.workflow_run_id
            organization_id = context.organization_id
            original_value = value
            value = get_actual_value_of_parameter_if_secret(workflow_run_id, original_value)

            # support TOTP secret and internal it to TOTP code
            is_totp_value = value == "BW_TOTP" or value == "OP_TOTP" or value == "AZ_TOTP"
            if is_totp_value:
                value = generate_totp_value(context.workflow_run_id, original_value)
            elif (totp_identifier or totp_url) and organization_id:
                # Try credential TOTP first (higher priority than webhook/totp_identifier)
                credential_totp = try_generate_totp_from_credential(workflow_run_id)
                if credential_totp:
                    value = credential_totp.value
                else:
                    totp_value = await poll_otp_value(
                        organization_id=organization_id,
                        task_id=task_id,
                        workflow_run_id=workflow_run_id,
                        totp_verification_url=totp_url,
                        totp_identifier=totp_identifier,
                    )
                    if totp_value:
                        # use the totp verification code
                        value = totp_value.value

        return value

    # Class-level cache for TOTP codes to ensure all digits in a sequence use the same code
    # Key: (workflow_run_id, credential_key), Value: totp_code
    # Uses TTLCache with 30-second expiry (aligned with TOTP rotation period)
    # and max 100 entries to prevent unbounded memory growth
    _totp_sequence_cache: TTLCache[tuple[str, str], str] = TTLCache(maxsize=100, ttl=30)

    async def get_totp_digit(
        self,
        context: Any,
        field_name: str,
        digit_index: int,
        totp_identifier: str | None = None,
        totp_url: str | None = None,
    ) -> str:
        """
        Get a specific digit from a TOTP code for multi-field TOTP inputs.

        This method is used by generated scripts for multi-field TOTP where each
        input field needs a single digit. It resolves the full TOTP code from
        the credential and returns the specific digit.

        IMPORTANT: When digit_index == 0, a fresh TOTP code is generated and cached.
        For digit_index > 0, the cached code is used. This ensures all 6 digits
        of a multi-field TOTP use the same code even if filling spans TOTP rotation
        boundaries.

        Args:
            context: The run context containing parameters
            field_name: The parameter name containing the TOTP code or credential reference
            digit_index: The index of the digit to return (0-5 for a 6-digit TOTP)
            totp_identifier: Optional TOTP identifier for polling
            totp_url: Optional TOTP verification URL

        Returns:
            The single digit at the specified index
        """
        totp_code = ""
        skyvern_ctx = skyvern_context.ensure_context()
        workflow_run_id = skyvern_ctx.workflow_run_id if skyvern_ctx else None

        LOG.info(
            "get_totp_digit called",
            field_name=field_name,
            digit_index=digit_index,
            workflow_run_id=workflow_run_id,
        )

        # Get the raw parameter value (may be credential reference like BW_TOTP)
        raw_value = context.parameters.get(field_name, "")

        # If the direct field_name parameter is empty, try to find a credential TOTP
        # by looking at the workflow run context for credential parameters
        if not raw_value and skyvern_ctx and workflow_run_id:
            workflow_run_context = app.WORKFLOW_CONTEXT_MANAGER.get_workflow_run_context(workflow_run_id)
            if workflow_run_context:
                # Look for credential parameters in the workflow run context values
                for key, value in workflow_run_context.values.items():
                    if key.startswith("cred_") and isinstance(value, dict) and "totp" in value:
                        cache_key = (workflow_run_id, key)

                        # For digit_index == 0, clear any stale cache and generate fresh TOTP
                        # For digit_index > 0, use cached code if available
                        if digit_index == 0:
                            # Clear stale cache for new sequence, fall through to generate
                            if cache_key in self._totp_sequence_cache:
                                del self._totp_sequence_cache[cache_key]
                        elif cache_key in self._totp_sequence_cache:
                            # Use cached value for digit_index > 0
                            totp_code = self._totp_sequence_cache[cache_key]
                            LOG.info(
                                "Using cached TOTP code for sequence",
                                field_name=field_name,
                                credential_key=key,
                                digit_index=digit_index,
                                totp_code_length=len(totp_code),
                            )
                            break

                        # Generate new TOTP code (digit_index==0 or cache miss)
                        totp_secret_id = value.get("totp")
                        if totp_secret_id:
                            totp_secret_key = workflow_run_context.totp_secret_value_key(totp_secret_id)
                            totp_secret = workflow_run_context.get_original_secret_value_or_none(totp_secret_key)
                            if totp_secret:
                                try:
                                    totp_code = pyotp.TOTP(totp_secret).now()
                                    # Cache the code for subsequent digit requests in this sequence
                                    self._totp_sequence_cache[cache_key] = totp_code
                                    LOG.info(
                                        "Generated fresh TOTP and cached for sequence",
                                        field_name=field_name,
                                        credential_key=key,
                                        digit_index=digit_index,
                                        totp_code_length=len(totp_code),
                                    )
                                    break
                                except Exception as e:
                                    LOG.warning(
                                        "Failed to generate TOTP code",
                                        credential_key=key,
                                        error=str(e),
                                    )

        # If we still don't have a TOTP code, try resolving via get_actual_value
        if not totp_code:
            totp_code = await self.get_actual_value(raw_value, totp_identifier, totp_url)

        # Return the specific digit
        if digit_index < len(totp_code):
            return totp_code[digit_index]
        LOG.warning(
            "TOTP digit index out of range",
            field_name=field_name,
            digit_index=digit_index,
            totp_code_length=len(totp_code),
        )
        return ""

    async def goto(self, url: str, **kwargs: Any) -> None:
        url = render_template(url)
        url = prepend_scheme_and_validate_url(url)

        # Print navigation in script mode
        context = skyvern_context.current()
        if context and context.script_mode:
            print(f"🌐 Navigating to: {url}")

        timeout = kwargs.pop("timeout", settings.BROWSER_LOADING_TIMEOUT_MS)
        await self.page.goto(url, timeout=timeout, **kwargs)

        if context and context.script_mode:
            print("  ✓ Page loaded")

    @action_wrap(ActionType.SOLVE_CAPTCHA)
    async def solve_captcha(
        self, prompt: str | None = None, data: str | dict[str, Any] | None = None, intention: str | None = None
    ) -> None:
        context = skyvern_context.current()
        if not context or not context.organization_id or not context.task_id or not context.step_id:
            await asyncio.sleep(30)
            return None

        task = await app.DATABASE.get_task(context.task_id, context.organization_id)
        step = await app.DATABASE.get_step(context.step_id, context.organization_id)
        if task and step:
            solve_captcha_handler = ActionHandler._handled_action_types[ActionType.SOLVE_CAPTCHA]
            action = SolveCaptchaAction(
                organization_id=context.organization_id,
                task_id=context.task_id,
                step_id=context.step_id,
            )
            await solve_captcha_handler(action, self.page, self.scraped_page, task, step)
        else:
            await asyncio.sleep(30)

    @action_wrap(ActionType.COMPLETE)
    async def complete(
        self, prompt: str | None = None, data: str | dict[str, Any] | None = None, intention: str | None = None
    ) -> None:
        # TODO: add validation here. if it doesn't pass the validation criteria:
        #  1. terminate the workflow run if fallback to ai is false
        #  2. fallback to ai if fallback to ai is true
        context = skyvern_context.current()
        if (
            not context
            or not context.organization_id
            or not context.workflow_run_id
            or not context.task_id
            or not context.step_id
        ):
            return
        task = await app.DATABASE.get_task(context.task_id, context.organization_id)
        step = await app.DATABASE.get_step(context.step_id, context.organization_id)
        if task and step:
            # CRITICAL: Update step.output with actions_and_results BEFORE validation
            # This ensures complete_verify() can access action history (including download info)
            # when checking if the goal was achieved
            await self._update_step_output_before_complete(context)
            # Refresh step to get updated output for validation
            step = await app.DATABASE.get_step(context.step_id, context.organization_id)
            if not step:
                return

            action = CompleteAction(
                organization_id=context.organization_id,
                task_id=context.task_id,
                step_id=context.step_id,
                step_order=step.order,
                action_order=context.action_order,
            )
            # result = await ActionHandler.handle_action(self.scraped_page, task, step, self.page, action)
            result = await handle_complete_action(action, self.page, self.scraped_page, task, step)
            if result and result[-1].success is False:
                raise ScriptTerminationException(result[-1].exception_message)

    async def _update_step_output_before_complete(self, context: skyvern_context.SkyvernContext) -> None:
        """Update step.output with actions_and_results before complete validation.

        This is critical for cached runs because complete_verify() reads action history
        from step.output.actions_and_results to check if goals were achieved (e.g., file downloads).
        Without this, the validation has no visibility into what actions were performed.
        """

        # Validate required context fields
        if not context.step_id or not context.task_id or not context.organization_id:
            return

        run_context = script_run_context_manager.get_run_context()
        if not run_context or not run_context.actions_and_results:
            return

        # Extract errors from DecisiveActions (similar to agent flow)
        errors: list[UserDefinedError] = []
        for action, _ in run_context.actions_and_results:
            if isinstance(action, DecisiveAction):
                errors.extend(action.errors)

        # Create AgentStepOutput similar to how agent does it
        step_output = AgentStepOutput(
            actions_and_results=run_context.actions_and_results,
            action_results=[result for _, results in run_context.actions_and_results for result in results],
            errors=errors,
        )

        await app.DATABASE.update_step(
            step_id=context.step_id,
            task_id=context.task_id,
            organization_id=context.organization_id,
            output=step_output,
        )
        LOG.info(
            "Updated step output with cached actions before complete validation",
            step_id=context.step_id,
            task_id=context.task_id,
            num_actions=len(run_context.actions_and_results),
        )


class ScriptRunContextManager:
    """
    Manages the run context for code runs.
    """

    def __init__(self) -> None:
        # self.run_contexts: dict[str, RunContext] = {}
        self.run_context: RunContext | None = None
        self.cached_fns: dict[str, Callable] = {}

    def get_run_context(self) -> RunContext | None:
        return self.run_context

    def set_run_context(self, run_context: RunContext) -> None:
        self.run_context = run_context

    def ensure_run_context(self) -> RunContext:
        if not self.run_context:
            raise Exception("Run context not found")
        return self.run_context

    def set_cached_fn(self, cache_key: str, fn: Callable) -> None:
        self.cached_fns[cache_key] = fn

    def get_cached_fn(self, cache_key: str | None = None) -> Callable | None:
        if cache_key:
            return self.cached_fns.get(cache_key)
        return None


script_run_context_manager = ScriptRunContextManager()



from __future__ import annotations

import asyncio
import os
from collections import OrderedDict
from contextvars import ContextVar, Token
from threading import RLock

import structlog

from skyvern.client import SkyvernEnvironment
from skyvern.config import settings
from skyvern.library.skyvern import Skyvern

from .api_key_hash import hash_api_key_for_cache

_skyvern_instance: ContextVar[Skyvern | None] = ContextVar("skyvern_instance", default=None)
_api_key_override: ContextVar[str | None] = ContextVar("skyvern_api_key_override", default=None)
_global_skyvern_instance: Skyvern | None = None
_api_key_clients: OrderedDict[str, Skyvern] = OrderedDict()
_clients_lock = RLock()
LOG = structlog.get_logger(__name__)


def _resolve_api_key_cache_size() -> int:
    raw = os.environ.get("SKYVERN_MCP_API_KEY_CLIENT_CACHE_SIZE", "128")
    try:
        return max(1, int(raw))
    except ValueError:
        return 128


_API_KEY_CLIENT_CACHE_MAX = _resolve_api_key_cache_size()


def _cache_key(api_key: str) -> str:
    """Hash API key so raw secrets are never stored as dict keys."""
    return hash_api_key_for_cache(api_key)


def _resolve_api_key() -> str | None:
    return settings.SKYVERN_API_KEY or os.environ.get("SKYVERN_API_KEY")


def _resolve_base_url() -> str | None:
    return settings.SKYVERN_BASE_URL or os.environ.get("SKYVERN_BASE_URL")


def _build_cloud_client(api_key: str) -> Skyvern:
    return Skyvern(
        api_key=api_key,
        environment=SkyvernEnvironment.CLOUD,
        base_url=_resolve_base_url(),
    )


def _close_skyvern_instance_best_effort(instance: Skyvern) -> None:
    """Close a Skyvern instance, regardless of whether an event loop is running."""
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        try:
            asyncio.run(instance.aclose())
        except Exception:
            LOG.debug("Failed to close evicted Skyvern client", exc_info=True)
        return

    task = loop.create_task(instance.aclose())

    def _on_done(done: asyncio.Task[None]) -> None:
        try:
            done.result()
        except Exception:
            LOG.debug("Failed to close evicted Skyvern client", exc_info=True)

    task.add_done_callback(_on_done)


def get_active_api_key() -> str | None:
    """Return the effective API key for this request/context."""
    return _api_key_override.get() or _resolve_api_key()


def set_api_key_override(api_key: str | None) -> Token[str | None]:
    """Set request-scoped API key override for MCP HTTP requests."""
    _skyvern_instance.set(None)
    return _api_key_override.set(api_key)


def reset_api_key_override(token: Token[str | None]) -> None:
    """Reset request-scoped API key override."""
    _api_key_override.reset(token)
    _skyvern_instance.set(None)


def get_skyvern() -> Skyvern:
    """Get or create a Skyvern client instance."""
    global _global_skyvern_instance

    override_api_key = _api_key_override.get()
    if override_api_key:
        instance = _skyvern_instance.get()
        if instance is None:
            key = _cache_key(override_api_key)
            evicted_clients: list[Skyvern] = []
            # Hold lock across lookup + build + insert to prevent two coroutines
            # from both building a client for the same API key concurrently.
            with _clients_lock:
                instance = _api_key_clients.get(key)
                if instance is not None:
                    _api_key_clients.move_to_end(key)
                else:
                    instance = _build_cloud_client(override_api_key)
                    _api_key_clients[key] = instance
                    _api_key_clients.move_to_end(key)
                    while len(_api_key_clients) > _API_KEY_CLIENT_CACHE_MAX:
                        _, evicted = _api_key_clients.popitem(last=False)
                        evicted_clients.append(evicted)
            for evicted in evicted_clients:
                _close_skyvern_instance_best_effort(evicted)
        _skyvern_instance.set(instance)
        return instance

    instance = _skyvern_instance.get()
    if instance is None:
        with _clients_lock:
            instance = _global_skyvern_instance
            if instance is None:
                api_key = _resolve_api_key()
                if api_key:
                    instance = _build_cloud_client(api_key)
                else:
                    instance = Skyvern.local()
                _global_skyvern_instance = instance
    _skyvern_instance.set(instance)
    return instance


async def close_skyvern() -> None:
    """Close active Skyvern client(s) and release Playwright resources."""
    global _global_skyvern_instance

    instances: list[Skyvern] = []
    seen: set[int] = set()
    with _clients_lock:
        candidates = (_skyvern_instance.get(), _global_skyvern_instance, *_api_key_clients.values())
        _api_key_clients.clear()
        _global_skyvern_instance = None

    for candidate in candidates:
        if candidate is None or id(candidate) in seen:
            continue
        seen.add(id(candidate))
        instances.append(candidate)

    for instance in instances:
        try:
            await instance.aclose()
        except Exception:
            LOG.warning("Failed to close Skyvern client", exc_info=True)

    _skyvern_instance.set(None)



"""db script for 1password integration

Revision ID: 1517a4ba63fa
Revises: add_run_timestamps
Create Date: 2025-06-12 08:06:13.439802+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "1517a4ba63fa"
down_revision: Union[str, None] = "add_run_timestamps"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "onepassword_credential_parameters",
        sa.Column("onepassword_credential_parameter_id", sa.String(), nullable=False),
        sa.Column("workflow_id", sa.String(), nullable=False),
        sa.Column("key", sa.String(), nullable=False),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("vault_id", sa.String(), nullable=False),
        sa.Column("item_id", sa.String(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=False),
        sa.Column("modified_at", sa.DateTime(), nullable=False),
        sa.Column("deleted_at", sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint("onepassword_credential_parameter_id"),
    )
    op.create_index(
        op.f("ix_onepassword_credential_parameters_onepassword_credential_parameter_id"),
        "onepassword_credential_parameters",
        ["onepassword_credential_parameter_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_onepassword_credential_parameters_workflow_id"),
        "onepassword_credential_parameters",
        ["workflow_id"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(
        op.f("ix_onepassword_credential_parameters_workflow_id"), table_name="onepassword_credential_parameters"
    )
    op.drop_index(
        op.f("ix_onepassword_credential_parameters_onepassword_credential_parameter_id"),
        table_name="onepassword_credential_parameters",
    )
    op.drop_table("onepassword_credential_parameters")
    # ### end Alembic commands ###



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .file_encoding import FileEncoding


class ScriptFileCreate(UniversalBaseModel):
    """
    Model representing a file in a script.
    """

    path: str = pydantic.Field()
    """
    File path relative to script root
    """

    content: str = pydantic.Field()
    """
    Base64 encoded file content
    """

    encoding: typing.Optional[FileEncoding] = pydantic.Field(default=None)
    """
    Content encoding
    """

    mime_type: typing.Optional[str] = pydantic.Field(default=None)
    """
    MIME type (auto-detected if not provided)
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



"""add run_signature to script block

Revision ID: b80c42316c94
Revises: 774e10939484
Create Date: 2025-10-14 22:32:57.466000+00:00

"""

from typing import Sequence, Union

import sqlalchemy as sa

from alembic import op

# revision identifiers, used by Alembic.
revision: str = "b80c42316c94"
down_revision: Union[str, None] = "774e10939484"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("script_blocks", sa.Column("run_signature", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("script_blocks", "run_signature")
    # ### end Alembic commands ###
